/*
  adc.cpp - ADC Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "adc.h"

AdcConfig::SAMPLES Adc::_samples;
Adc::MODE Adc::_mode;
uint32    Adc::_start;
uint16    Adc::_periods;
uint16    Adc::_adcref;
uint16    Adc::_adcsum;
uint16    Adc::_result;
uint8     Adc::_count;
uint8     Adc::_status;

uint8     MAdc::_count;
uint8     MAdc::_index;
AdcConfig MAdc::_config[];
uint16    MAdc::_result[];
uint16    MAdc::_adcref[];

AdcConfig& AdcConfig::source(SRC source)
{
  _source  = source;
  _range   = RANGE_1;
  _samples = SAMPLES_1;
  _adcref  = ANALOG_BANDGAP;
  _clock   = Analog::CLOCK_500KHZ;
  _capture = Analog::CAPTURE_2;
  return *this;
}

AdcConfig& AdcConfig::range(RANGE value)
{
  _range = value;
  return *this;
}

AdcConfig& AdcConfig::samples(SAMPLES value)
{
  _samples = value;
  return *this;
}

AdcConfig& AdcConfig::reference(uint16 value)
{
  _adcref = value;
  return *this;
}

AdcConfig& AdcConfig::clock(Analog::CLOCK value)
{
  _clock = value;
  return *this;
}

AdcConfig& AdcConfig::capture(Analog::CAPTURE value)
{
  _capture = value;
  return *this;
}

void AdcConfig::configure(void)
{
  Analog::clock(_clock);
  Analog::capture(_capture);
  Analog::reference(_adcref != ANALOG_BANDGAP);
  Analog::interrupt(false);
  Analog::configure();
}

uint16 AdcConfig::periods(void)
{
  // return 16MHz clock periods.
  return (((_capture + 1) << 1) * 3 + 13) << (_clock + 3);
}

void Adc::begin(MODE mode, AdcConfig& conf)
{
  AdcConfig::RANGE range = conf._range;
  _adcref = conf._adcref;
  if (conf._adcref == ANALOG_BANDGAP)
  {
    if (conf._source == AdcConfig::SRC_ADC_TEMP)
      range = AdcConfig::RANGE_1;
    else if (conf._source == AdcConfig::SRC_ADC_VOLT)
    {
      range = AdcConfig::RANGE_2;
      _adcref = ANALOG_BANDGAP * 2 * 1000 / 666;
    }
  }
  else if (conf._range)
    _adcref <<= 1;
  switch ((uint8)conf._source)
  {
    case AdcConfig::SRC_ADC_3: Gpio::pinMode(Gpio::DIO0, Gpio::INPUT); break;
    case AdcConfig::SRC_ADC_4: Gpio::pinMode(Gpio::DIO1, Gpio::INPUT); break;
  }
  _samples = conf._samples;
  _periods = conf.periods();
  _mode    = mode;
  _status  = 0;
  _adcsum  = 0;
  _count   = 0;
  _result  = 0;
  conf.configure();
  vAHI_AdcEnable(E_AHI_ADC_SINGLE_SHOT, range, conf._source);
}

void Adc::end(void)
{
  if (_status)
  {
    _status = 0;
    vAHI_AdcDisable();
  }
}

void Adc::start(void)
{
  if (!(_status & ~COMVERSION_END))
  {
    _status = 1 << _mode;
    _adcsum = 0;
    _count  = 0;
    vAHI_AdcStartSample();
    _start  = TickTimer::read();
  }
}

bool Adc::conversion(void)
{
  if (_status & COMVERSION_END)
    return true;
  if (!_status || (TickTimer::read() - _start < _periods))
    return false;
  uint16 value = u16AHI_AdcRead();
  if (_samples >= AdcConfig::SAMPLES_1)
    _result = value;
  else
  {
    _adcsum += value;
    if (++_count < (1 << (_samples + 1)))
    {
      vAHI_AdcStartSample();
      _start = TickTimer::read();
      return false;
    }
    _result = (_adcsum + (1 << _samples)) >> (_samples + 1);
    _adcsum = 0;
    _count  = 0;
  }
  if (_mode == MODE_CONTINUOUS)
  {
    vAHI_AdcStartSample();
    _start  = TickTimer::read();
    _status = COMVERSION_END | (1 << MODE_CONTINUOUS);
  }
  else
    _status = COMVERSION_END;
  return true;
}

uint16 Adc::read(void)
{
  _status &= ~COMVERSION_END;
  return _result;
}

uint16 Adc::voltage(void)
{
  return read() * _adcref / ADC_MAX_VALUE;
}

int16 Adc::temperature(void)
{
  return temperature(voltage());
}

int16 Adc::temperature(int16 volt)
{
  if (volt <= 610)
    return -40;
  if (volt >= 840)
    return 125;
  volt -= 730;
  return ADC_TEMP_CORRECTION_VALUE + 25 + (volt < 0 ? volt * 65 / 120 : volt * 100 / 110);
}

uint16 Adc::reference(void)
{
  return _adcref;
}

void MAdc::begin(const AdcConfig *conf, uint8 size)
{
  _index = 0;
  _count = 0;
  while (size-- && (_count < ADC_SRC_MAX))
    _config[_count++] = *conf++;
  if (_count)
    next(false);
}

void MAdc::end(void)
{
  _count = 0;
  Adc::end();
}

uint16 MAdc::read(AdcConfig::SRC source)
{
  return _result[source];
}

uint16 MAdc::voltage(AdcConfig::SRC source)
{
  return read(source) * _adcref[source] / ADC_MAX_VALUE;
}

int16 MAdc::temperature(void)
{
  return Adc::temperature(voltage(AdcConfig::SRC_ADC_TEMP));
}

void MAdc::handle(void)
{
  if (_count)
  {
    if (Adc::conversion())
      next(true);
  }
}

void MAdc::next(bool conversion)
{
  if (conversion)
  {
    AdcConfig::SRC source = _config[_index]._source;
    _result[source] = Adc::read();
    _adcref[source] = Adc::reference();
    if (++_index >= _count)
      _index = 0;
  }
  Adc::begin(Adc::MODE_SINGLE_SHOT, _config[_index]);
  Adc::start();
}
