/*
  adc.h - ADC Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <jendefs.h>
#include <AppHardwareApi.h>
#include "ticktimer.h"
#include "analog.h"
#include "gpio.h"

#define ADC_BIT_RESOLUTION 10 // bits
#define ADC_MAX_VALUE      ((1 << ADC_BIT_RESOLUTION) - 1)

#ifndef ADC_TEMP_CORRECTION_VALUE
#define ADC_TEMP_CORRECTION_VALUE 14
#endif

/*
   I/O Ports

     ADC1: ADC1
     ADC2: ADC2_VREF
     ADC3: DIO0
     ADC4: DIO1

   The following SDK API seems to have a bug.

     vAHI_AdcStartAccumulateSamples()
     bAHI_AdcPoll()
     bAHI_AdcEnableSampleBuffer()
*/

class AdcConfig
{
  public:

    typedef enum
    {
      SRC_ADC_1    = E_AHI_ADC_SRC_ADC_1, // ADC1 input
      SRC_ADC_2    = E_AHI_ADC_SRC_ADC_2, // ADC2 input
      SRC_ADC_3    = E_AHI_ADC_SRC_ADC_3, // ADC3 input
      SRC_ADC_4    = E_AHI_ADC_SRC_ADC_4, // ADC4 input
#ifdef E_AHI_ADC_SRC_ADC_5
      SRC_ADC_5    = E_AHI_ADC_SRC_ADC_5, // ADC5 input on JN5169
#endif
#ifdef E_AHI_ADC_SRC_ADC_6
      SRC_ADC_6    = E_AHI_ADC_SRC_ADC_6, // ADC6 input on JN5169
#endif
      SRC_ADC_TEMP = E_AHI_ADC_SRC_TEMP,  // on-chip temperature sensor
      SRC_ADC_VOLT = E_AHI_ADC_SRC_VOLT,  // internal voltage monitor
    } SRC;

    typedef enum
    {
      RANGE_1 = E_AHI_AP_INPUT_RANGE_1, // 0 to Vref
      RANGE_2 = E_AHI_AP_INPUT_RANGE_2, // 0 to 2Vref
    } RANGE;

    typedef enum
    {
      SAMPLES_2  = E_AHI_ADC_ACC_SAMPLE_2,  //  2 samples
      SAMPLES_4  = E_AHI_ADC_ACC_SAMPLE_4,  //  4 samples
      SAMPLES_8  = E_AHI_ADC_ACC_SAMPLE_8,  //  8 samples
      SAMPLES_16 = E_AHI_ADC_ACC_SAMPLE_16, // 16 samples
      SAMPLES_1,
    } SAMPLES;

    /* must be first call */
    AdcConfig& source(SRC source);
    AdcConfig& range(RANGE value);
    AdcConfig& samples(SAMPLES value);
    AdcConfig& reference(uint16 value);
    AdcConfig& clock(Analog::CLOCK value);
    AdcConfig& capture(Analog::CAPTURE value);

  private:

    void configure(void);
    uint16 periods(void);

    friend class Adc;
    friend class MAdc;

    SRC             _source;
    RANGE           _range;
    SAMPLES         _samples;
    uint16          _adcref;
    Analog::CLOCK   _clock;
    Analog::CAPTURE _capture;
};

class Adc
{
  public:

    typedef enum
    {
      MODE_SINGLE_SHOT = E_AHI_ADC_SINGLE_SHOT, // single-shot mode
      MODE_CONTINUOUS  = E_AHI_ADC_CONTINUOUS,  // continous mode
    } MODE;

    static void begin(MODE mode, AdcConfig& conf);
    static void end(void);
    static void start(void);
    static bool conversion(void);
    static uint16 read(void);
    static uint16 voltage(void);
    static int16 temperature(void);
    static int16 temperature(int16 volt);
    static uint16 reference(void);

  private:

    static const uint8 COMVERSION_END = 0x80;

    static AdcConfig::SAMPLES _samples;
    static MODE   _mode;
    static uint32 _start;
    static uint16 _periods;
    static uint16 _adcref;
    static uint16 _adcsum;
    static uint16 _result;
    static uint8  _count;
    static uint8  _status;
};

class MAdc
{
  public:

    static void begin(const AdcConfig *conf, uint8 size);
    static void end(void);
    static uint16 read(AdcConfig::SRC source);
    static uint16 voltage(AdcConfig::SRC source);
    static int16 temperature(void);
    static void handle(void);

  private:

    static void next(bool conversion);

    static const uint8 ADC_SRC_MAX = AdcConfig::SRC_ADC_VOLT + 1;

    static uint8     _index;
    static uint8     _count;
    static AdcConfig _config[ADC_SRC_MAX];
    static uint16    _result[ADC_SRC_MAX];
    static uint16    _adcref[ADC_SRC_MAX];
};
