/*
  alarm.cpp - Alarm Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "system.h"
#include "alarm.h"

#define lengthof(a) (sizeof(a)/sizeof(a[0]))

AlarmClass AlarmTimer;

static AlarmClass *_instances[5];

void AlarmClass::begin(void)
{
  _timer = (Timer::TIMER)-1;
#ifdef ALARM_TIMER
  begin((Timer::TIMER)ALARM_TIMER);
#endif
}

void AlarmClass::begin(Timer::TIMER timer)
{
  if (timer < lengthof(_instances))
  {
    _timer      = timer;
    _alarm_list = NULL;
    _instances[_timer] = this;
    Timer::begin(timer, Timer::CLOCK_1MHZ, false, true, false, TimerCallback);
  }
}

bool AlarmClass::start(Alarm& alarm)
{
  if (((int8)_timer < 0) || (alarm._interval < 1) || !alarm._handler)
    return false;
  uint32 save = System::disableAndSaveInterrupts();
  if (add(alarm))
    restart(_alarm_list->_alarm - System::micros64());
  System::restoreInterrupts(save);
  return true;
}

void AlarmClass::cancel(Alarm& alarm)
{
  uint32 save = System::disableAndSaveInterrupts();
  for (Alarm **p = &_alarm_list; *p; p = &(*p)->_next)
  {
    if (*p == &alarm)
    {
      *p = (*p)->_next;
      break;
    }
  }
  System::restoreInterrupts(save);
}

void AlarmClass::cancel(uint8 num, void *arg)
{
  if ((num < lengthof(Alarm::_params)) && arg)
  {
    uint32 save = System::disableAndSaveInterrupts();
    for (Alarm **p = &_alarm_list; *p; p = &(*p)->_next)
    {
      if ((*p)->_params[num] == arg)
      {
        *p = (*p)->_next;
        break;
      }
    }
    System::restoreInterrupts(save);
  }
}

bool AlarmClass::add(Alarm& alarm)
{
  Alarm **p;
  for (p = &_alarm_list; *p && ((int64)((*p)->_alarm - alarm._alarm) <= 0); p = &(*p)->_next)
    continue;
  alarm._next = *p;
  *p = &alarm;
  return p == &_alarm_list;
}

void AlarmClass::TimerCallback(uint32 u32Device, uint32 u32ItemBitmap)
{
  (void)u32Device;     // E_AHI_DEVICE_TIMERx
  (void)u32ItemBitmap; // E_AHI_TIMER_PERIOD_MASK or E_AHI_TIMER_RISE_MASK
  _instances[Timer::id(u32Device)]->timeup();
}

void AlarmClass::timeup(void)
{
  Alarm *p = NULL;
  while (_alarm_list)
  {
    int64 t = _alarm_list->_alarm - System::micros64();
    if (p || (t > 0))
    {
      restart(t);
      return;
    }
    _alarm_list = (p = _alarm_list)->_next;
    if (p->_handler(*p))
    {
      p->_alarm += p->_interval;
      if (!add(*p))
        p = NULL;
    }
    else
      p = NULL;
  }
}

void AlarmClass::restart(int64 interval)
{
  uint16 n;
  if (interval <= 1)
    n = 2;
  else if (interval & ~0xFFFF)
    n = (uint16)-1000; // leave 1ms or more.
  else
    n = (uint16)interval;
  Timer::startSingleShot(_timer, 0, n);
}

Alarm& Alarm::interval(int64 interval, bool fromnow)
{
  if (fromnow)
    _alarm = System::micros64() + interval;
  _interval = interval;
  return *this;
}

int64 Alarm::interval(void)
{
  return _interval;
}

Alarm& Alarm::handler(Alarm::handler_t handler)
{
  _handler = handler;
  return *this;
}

Alarm::handler_t Alarm::handler(void)
{
  return _handler;
}

Alarm& Alarm::param(uint8 num, void *val)
{
  if (num < lengthof(_params))
    _params[num] = val;
  return *this;
}

void *Alarm::param(uint8 num)
{
  return num < lengthof(_params) ? _params[num] : NULL;
}

bool Alarm::expire(void)
{
  if ((_interval >> 32) || ((int32)_interval < 0))
  {
    if ((int64)(System::micros64() - _alarm) < 0)
      return false;
  }
  else
  {
    if ((int32)(System::micros() - (uint32)_alarm) < 0)
      return false;
  }
  _alarm += _interval;
  return true;
}
