/*
  twelite_cue.h - Board Library for MONOWIRESS TWELITE CUE

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include "gpio.h"
#include "led.h"
#include "alarm.h"
#include "mc3630_jn516x.h"  // MC3630_IO

class Board
{
  public:

    static const Gpio::PIN PIN_LED      = Gpio::DIO5;
    static const Gpio::PIN PIN_WDTOUT   = Gpio::DIO13;
    static const Gpio::PIN PIN_SNS_INT  = Gpio::DIO17;
    static const Gpio::PIN PIN_SNS_OUT1 = Gpio::DIO16;
    static const Gpio::PIN PIN_SNS_OUT2 = Gpio::DIO8;
    static const Gpio::PIN PIN_SW_SET   = Gpio::DIO1;

    static
    void begin(void)
    {
      /* Setup LED */
      Led::setup(0, PIN_LED, HIGH);
      /* Setup GPIO */
      Gpio::pinMode(PIN_WDTOUT  , Gpio::OUTPUT);
      Gpio::pinMode(PIN_SW_SET  , Gpio::INPUT_PULLUP);
      Gpio::pinMode(PIN_SNS_OUT1, Gpio::INPUT);
      Gpio::pinMode(PIN_SNS_OUT2, Gpio::INPUT);
      Gpio::pinMode(PIN_SNS_INT , Gpio::INPUT_PULLUP);
      Gpio::wake(PIN_SNS_INT, Gpio::FALLING, empty);
      /* Setup MC3630 */
      _mc3630.begin(MC3630::IFMODE_SPI, MC3630_IO::begin(MC3630::IFMODE_SPI));
      _mc3630.setInterrupt(MC3630::INTE_WAKE);
      _mc3630.setWakeConfig(MC3630::RATE_ULP_SLOW, MC3630::RANGE_12G, MC3630::RES_12BIT);
      _mc3630.setSniffConfig(MC3630::SNIFF_RATE_ULP_DEFAULT);
      startCWake();
      /* shake parameter */
      _threshold = 1000;
      _mc3630.getOutput(_axis_sampl[_axis_index ^= 1]);
      _mc3630.getOutput(_axis_sampl[_axis_index ^= 1]);
      _alarm.interval(10000); // 10ms
    }

    static
    void startSniff(uint8 threshold, uint8 count = 0)
    {
      _mc3630.setMode(MC3630::MODE_STANDBY);
      _mc3630.clearInterrupt();
      _mc3630.setSniffThreshold(MC3630::THADR_THRESHOLD_X, threshold);
      _mc3630.setSniffThreshold(MC3630::THADR_THRESHOLD_Y, threshold);
      _mc3630.setSniffThreshold(MC3630::THADR_THRESHOLD_Z, threshold);
      _mc3630.setSniffThreshold(MC3630::THADR_COUNT_X, count);
      _mc3630.setSniffThreshold(MC3630::THADR_COUNT_Y, count);
      _mc3630.setSniffThreshold(MC3630::THADR_COUNT_Z, count);
      _mc3630.setMode(MC3630::MODE_SNIFF);
    }

    static
    void startCWake(void)
    {
      _mc3630.setMode(MC3630::MODE_STANDBY);
      _mc3630.clearInterrupt();
      _mc3630.setMode(MC3630::MODE_CWAKE);
    }

    static
    uint8_t readSW(void)
    {
      return Gpio::digitalRead(PIN_SW_SET);
    }

    static
    uint8_t readMagneticSensor(void)
    {
      return (Gpio::digitalRead(PIN_SNS_OUT1) << 0) | (Gpio::digitalRead(PIN_SNS_OUT2) << 1);
    }

    static
    MC3630& getAccelerometer(void)
    {
      return _mc3630;
    }

    static
    void handle(void)
    {
      Gpio::digitalWrite(PIN_WDTOUT, HIGH);
      Gpio::digitalWrite(PIN_WDTOUT, LOW);
      if (_alarm.expire())
      {
        _mc3630.getOutput(_axis_sampl[_axis_index ^= 1]);
        int rv = shake0();
        if (rv >= 0)
          _axis_shake |= 1 << rv;
      }
    }

    static
    uint8_t dice(void)
    {
      int x = _axis_sampl[0][MC3630::AXIS_X];
      int y = _axis_sampl[0][MC3630::AXIS_Y];
      int z = _axis_sampl[0][MC3630::AXIS_Z];
      if (abs(z) > abs(x))
      {
        if (abs(z) > abs(y))
          return z < 0 ? 6 : 1;
      }
      else
      {
        if (abs(x) > abs(y))
          return x < 0 ? 3 : 4;
      }
      return y < 0 ? 5 : 2;
    }

    static
    uint8_t shake(void)
    {
      uint8_t rv = _axis_shake;
      _axis_shake = 0;
      return rv;
    }

    static
    void shake(uint16_t threshold)
    {
      _threshold = threshold;
    }

  private:

    static
    int abs(int val)
    {
      return val < 0 ? - val : val;
    }

    static
    int shake0(void)
    {
      int x = abs(_axis_sampl[0][MC3630::AXIS_X] - _axis_sampl[1][MC3630::AXIS_X]);
      int y = abs(_axis_sampl[0][MC3630::AXIS_Y] - _axis_sampl[1][MC3630::AXIS_Y]);
      int z = abs(_axis_sampl[0][MC3630::AXIS_Z] - _axis_sampl[1][MC3630::AXIS_Z]);
      if (z > x)
      {
        if (z > y)
          return z >= _threshold ? MC3630::AXIS_Z : -1;
      }
      else
      {
        if (x > y)
          return x >= _threshold ? MC3630::AXIS_X : -1;
      }
      return y >= _threshold ? MC3630::AXIS_Y : -1;
    }

    static MC3630   _mc3630;
    static Alarm    _alarm;
    static int16_t  _axis_sampl[2][MC3630_NUMBER_OF_AXIS];
    static uint8_t  _axis_index;
    static uint8_t  _axis_shake;
    static uint16_t _threshold;
};
