/*
  debug.cpp - Debug Utility for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <ctype.h>
#include <string.h>
#include "debug.h"
#include "system.h"
#include "muldiv.h"

Uart *Debug::_output;
Mutex Debug::_mutex;

void Debug::begin(Uart::UART uart)
{
  switch (uart)
  {
    case Uart::UART_0 : _output = &Serial0; break;
    case Uart::UART_1 : _output = &Serial1; break;
  }
  if (_output)
    _output->begin();
  _mutex.init();
}

void Debug::mdump(const uint8 *mem, uint16 len)
{
  if (len)
  {
    _mutex.lock();
    int i = 0;
    printf("%02X", *mem++);
    while (++i < len)
      printf("%s%02X", i & 0x0F ? ((i & 0x0F) != 8 ? " " : "-") : "\r\n", *mem++);
    printf("\n");
    _mutex.unlock();
  }
}

void Debug::printf(const char *format, ...)
{
  const char *p = format;
  char c;
  va_list ap;

  va_start(ap, format);

  _mutex.lock();

  do
  {
    c = *p++;
    if (c == '%')
    {
      format_t option = {fillchar : ' '};
      if ((*p == '-') || (*p == '+'))
        option.alignchar = *p++;
      if (*p == '0')
        option.fillchar = *p++;
      while (isdigit(*p))
        option.width = option.width * 10 + (*p++ - '0');
      if (*p == '.')
      {
        ++p;
        while (isdigit(*p))
          option.precision = option.precision * 10 + (*p++ - '0');
      }
      while (*p == 'l')
      {
        ++p;
        ++option.typelong;
      }
      switch (*p++)
      {
        /* %d - integer */
        case 'd':
          option.value = option.typelong < 2 ? va_arg(ap, int32) : va_arg(ap, int64);
          if ((int64)option.value < 0)
          {
            option.value = 0 - option.value;
            option.signchar = '-';
          }
          else if (option.alignchar == '+')
            option.signchar = '+';
          option.base = 10;
          nformat(option);
          break;
        /* %u - unsigned integer */
        case 'u':
          option.value = option.typelong < 2 ? va_arg(ap, uint32) : va_arg(ap, uint64);
          option.base  = 10;
          nformat(option);
          break;
          /* %X - hex */
        case 'X':
          option.value = option.typelong < 2 ? va_arg(ap, uint32) : va_arg(ap, uint64);
          option.hexchar = 'A';
          option.base    = 16;
          nformat(option);
          break;
        /* %x - hex */
        case 'x':
          option.value = option.typelong < 2 ? va_arg(ap, uint32) : va_arg(ap, uint64);
          option.hexchar = 'a';
          option.base    = 16;
          nformat(option);
          break;
        /* %o - octal */
        case 'o':
          option.value = option.typelong < 2 ? va_arg(ap, uint32) : va_arg(ap, uint64);
          option.base  = 8;
          nformat(option);
          break;
        /* %b - binary */
        case 'b':
          option.value = option.typelong < 2 ? va_arg(ap, uint32) : va_arg(ap, uint64);
          option.base  = 2;
          nformat(option);
          break;
        /* %c - character */
        case 'c':
          putch(va_arg(ap, int));
          break;
        /* %s - string */
        case 's':
          option.fillchar = ' ';
          putstr(option, va_arg(ap, char *));
          break;
        /* %% - '%' character */
        case '%':
          putch('%');
          break;
        /* %something else not handled! */
        default:
          putch('?');
          break;
      }
    }
    else if (c)
    {
      if (c == '\n')
        putch('\r');
      putch(c);
    }
  }
  while (c);

  _mutex.unlock();

  va_end(ap);
}

void Debug::putch(char c)
{
  if (_output)
  {
    _mutex.lock();
    while (_output->write(c) == 0)
      yield();
    _mutex.unlock();
  }
}

void Debug::putstr(format_t& option, const char *str)
{
  if (!str)
    str = "(null)";
  size_t i, len = strlen(str);

  if (option.precision && (len > option.precision))
    len = option.precision;

  if (option.alignchar == '-')
  {
    for (i = 0; i < len; ++i)
      putch(*str++);
    for (i = len; i < option.width; ++i)
      putch(option.fillchar);
  }
  else
  {
    for (i = len; i < option.width; ++i)
      putch(option.fillchar);
    for (i = 0; i < len; ++i)
      putch(*str++);
  }
}

void Debug::nformat(format_t& option)
{
  char buf[128], *p = buf + sizeof(buf);
  *--p = 0;

  do
  {
    uint8 c;
    if (option.typelong < 2)
    {
      c = (uint32)option.value % option.base;
      option.value = (uint32)option.value / option.base;
    }
    else
    {
      uint64 r;
      option.value = MulDiv::divu64u64(option.value, option.base, &r);
      c = (uint8)r;
    }
    if (c < 10)
      *--p = '0' + c;
    else
      *--p = option.hexchar + (c - 10);
  }
  while (option.value);

  if (option.signchar)
    *--p = option.signchar;

  putstr(option, p);
}
