/*
  eeprom.cpp - EEPROM Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <string.h>
#include "eeprom.h"

uint8  EEPRom::_segbytes = 64;
uint16 EEPRom::_segments = 64;

void EEPRom::begin(void)
{
  _segments = u16AHI_InitialiseEEP(&_segbytes);
}

bool EEPRom::erase(uint16 u16Offset, uint16 u16Length)
{
  if (u16Offset > _segments * _segbytes)
    return false;
  uint16 u16SegmentIndex      = u16Offset / _segbytes;
  uint8  u8SegmentByteAddress = u16Offset % _segbytes;
  uint8  save[_segbytes];
  while (u16Length)
  {
    uint8 u8Len = MIN(u16Length, _segbytes - u8SegmentByteAddress);
    if (u8Len != _segbytes)
    {
      if (iAHI_ReadDataFromEEPROMsegment(u16SegmentIndex, 0, save, _segbytes))
        return false;
      memset(&save[u8SegmentByteAddress], 0x00, u8Len);
    }
    if (iAHI_EraseEEPROMsegment(u16SegmentIndex))
      return false;
    if (u8Len != _segbytes)
    {
      if (iAHI_WriteDataIntoEEPROMsegment(u16SegmentIndex, 0, save, _segbytes))
        return false;
    }
    u16SegmentIndex++;
    u8SegmentByteAddress = 0;
    u16Length -= u8Len;
  }
  return true;
}

bool EEPRom::write(uint16 u16Offset, uint8 *pu8Buffer, uint16 u16Length)
{
  if (u16Offset > _segments * _segbytes)
    return false;
  uint16 u16SegmentIndex      = u16Offset / _segbytes;
  uint8  u8SegmentByteAddress = u16Offset % _segbytes;
  while (u16Length)
  {
    uint8 u8Len = MIN(u16Length, _segbytes - u8SegmentByteAddress);
    if (iAHI_WriteDataIntoEEPROMsegment(u16SegmentIndex, u8SegmentByteAddress, pu8Buffer, u8Len))
      return false;
    u16SegmentIndex++;
    u8SegmentByteAddress = 0;
    u16Length -= u8Len;
    pu8Buffer += u8Len;
  }
  return true;
}

bool EEPRom::read(uint16 u16Offset, uint8 *pu8Buffer, uint16 u16Length)
{
  if (u16Offset > _segments * _segbytes)
    return false;
  uint16 u16SegmentIndex      = u16Offset / _segbytes;
  uint8  u8SegmentByteAddress = u16Offset % _segbytes;
  while (u16Length)
  {
    uint8 u8Len = MIN(u16Length, _segbytes - u8SegmentByteAddress);
    if (iAHI_ReadDataFromEEPROMsegment(u16SegmentIndex, u8SegmentByteAddress, pu8Buffer, u8Len))
      return false;
    u16SegmentIndex++;
    u8SegmentByteAddress = 0;
    u16Length -= u8Len;
    pu8Buffer += u8Len;
  }
  return true;
}
