/*
  led.cpp - LED Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "led.h"

Led::CTRL Led::_leds[];

#define LED_NONE (uint8)-1

void Led::begin(void)
{
  CTRL *ctrl = _leds;
  for (uint8 i = 0; i < LED_COUNT; ++i, ++ctrl)
  {
    ctrl->pin    = (uint8)-1;
    ctrl->off    = LOW;
    ctrl->mode   = MODE_DISABLE;
    ctrl->expire = false;
  }
}

void Led::setup(uint8 led, uint8 pin, uint8 off)
{
  CTRL& ctrl = _leds[led & LED_MASK];
  ctrl.pin    = pin;
  ctrl.off    = off;
  ctrl.mode   = MODE_DISABLE;
  ctrl.expire = false;
  Gpio::digitalWrite((Gpio::PIN)pin, off);
  Gpio::pinMode((Gpio::PIN)pin, Gpio::OUTPUT);
}

void Led::off(uint8 led, uint16 ms)
{
  control(_leds[led & LED_MASK], MODE_OFF, ms);
}

void Led::on(uint8 led, uint16 ms)
{
  control(_leds[led & LED_MASK], MODE_ON, ms);
}

void Led::toggle(uint8 led, uint16 ms)
{
  control(_leds[led & LED_MASK], MODE_TOGGLE, ms);
}

void Led::control(CTRL& ctrl, MODE mode, uint16 ms)
{
  ctrl.expire = false;
  if (ms)
  {
    ctrl.mode = mode;
    ctrl.alarm.interval(ms * 1000UL);
  }
  else
  {
    ctrl.mode = MODE_DISABLE;
    if (ctrl.pin != LED_NONE)
    {
      switch (mode)
      {
        case MODE_OFF:    Gpio::digitalWrite((Gpio::PIN)ctrl.pin,  ctrl.off); break;
        case MODE_ON:     Gpio::digitalWrite((Gpio::PIN)ctrl.pin, !ctrl.off); break;
        case MODE_TOGGLE: Gpio::digitalWrite((Gpio::PIN)ctrl.pin, !Gpio::digitalRead((Gpio::PIN)ctrl.pin)); break;
        default: break;
      }
    }
  }
}

bool Led::status(uint8 led)
{
  CTRL& ctrl = _leds[led & LED_MASK];
  return Gpio::digitalRead((Gpio::PIN)ctrl.pin) != ctrl.off;
}

bool Led::expire(uint8 led)
{
  CTRL& ctrl = _leds[led & LED_MASK];
  bool rv = ctrl.expire;
  ctrl.expire = false;
  return rv;
}

void Led::handle(void)
{
  CTRL *ctrl = _leds;
  for (uint8 i = 0; i < LED_COUNT; ++i, ++ctrl)
  {
    MODE mode = ctrl->mode;
    if ((mode != MODE_DISABLE) && ctrl->alarm.expire())
    {
      control(*ctrl, mode, 0);
      ctrl->expire = true;
      if (mode == MODE_TOGGLE)
        ctrl->mode = mode;
    }
  }
}
