/*
  mc3630.cpp - 3-Axis Accelerometer Sensor Library for mCube - MC3630

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "system.h"
#include "mc3630.h"

bool MC3630::begin(IFMODE mode, IOMETHOD *iomethod)
{
  _iomethod = iomethod;
  initialize();
  if (mode & IFMODE_I2C)
  {
    setInterface(mode);
    setI2CWrClear(true); // Write clear interrupt
  }
  else
  {
    setInterface(mode);
    setSpiHsMode();      // Change SPI HS Mode (8MHz)
  }
  _registers.status_1.bit.FIFO_EMPTY = 1;
  return _enable = (getChipID() == 0x71);
}

MC3630::EXT_STAT_1 MC3630::getExtStatus1(void)
{
  if (_enable)
    readReg(&_registers.ext_stat_1, sizeof(_registers.ext_stat_1));
  return (EXT_STAT_1)_registers.ext_stat_1;
}

MC3630::EXT_STAT_2 MC3630::getExtStatus2(void)
{
  if (_enable)
    readReg(&_registers.ext_stat_2, sizeof(_registers.ext_stat_2));
  return (EXT_STAT_2)_registers.ext_stat_2;
}

MC3630::STATUS_1 MC3630::getStatus1(void)
{
  if (_enable)
    readReg(&_registers.status_1, sizeof(_registers.status_1));
  return (STATUS_1)_registers.status_1;
}

MC3630::STATUS_2 MC3630::getStatus2(void)
{
  if (_enable)
    readReg(&_registers.status_2, sizeof(_registers.status_2));
  return (STATUS_2)_registers.status_2;
}

void MC3630::getOutput(int16_t axis[MC3630_NUMBER_OF_AXIS])
{
  if (_enable)
  {
    STATUS_1 status = getStatus1();
    if (status.bit.NEW_DATA || !status.bit.FIFO_EMPTY)
      readReg(_registers.output, sizeof(_registers.output));
    for (uint8_t i = 0; i < MC3630_NUMBER_OF_AXIS; ++i)
      axis[i] = ((uint16_t)_registers.output[i].reg.H << 8) | _registers.output[i].reg.L;
  }
}

void MC3630::setOffset(const int16_t axis[MC3630_NUMBER_OF_AXIS])
{
  for (uint8_t i = 0; i < MC3630_NUMBER_OF_AXIS; ++i)
  {
    _registers.offset[i].bit.L = (uint8_t)(axis[i] >> 0);
    _registers.offset[i].bit.H = (uint8_t)(axis[i] >> 8);
  }
  writeReg(_registers.offset, sizeof(_registers.offset));
}

void MC3630::setGain(const int16_t axis[MC3630_NUMBER_OF_AXIS])
{
  for (uint8_t i = 0; i < MC3630_NUMBER_OF_AXIS; ++i)
  {
    uint16_t val = (axis[i] < 0x01FF ? axis[i] : 0x01FF);
    _registers.gain  [i]           = (uint8_t)(val >> 0);
    _registers.offset[i].bit.GAIN8 = (uint8_t)(val >> 8);
  }
  writeReg(_registers.offset, sizeof(_registers.offset) + sizeof(_registers.gain));
}

void MC3630::disableDriveMotion(const MOTION axis[MC3630_NUMBER_OF_AXIS])
{
  for (uint8_t i = 0; i < MC3630_NUMBER_OF_AXIS; ++i)
  {
    _registers.dm[i].bit.DP = (axis[i] & MOTION_POSITIVE ? 1 : 0);
    _registers.dm[i].bit.DN = (axis[i] & MOTION_NEGATIVE ? 1 : 0);
  }
  writeReg(_registers.dm, sizeof(_registers.dm));
}

void MC3630::setInterface(IFMODE mode)
{
  _registers.freg_1.bit.IFMODE = mode;
  writeReg(&_registers.freg_1, sizeof(_registers.freg_1));
}

MC3630::IFMODE MC3630::getInterface(void)
{
  readReg(&_registers.freg_1, sizeof(_registers.freg_1));
  return (IFMODE)_registers.freg_1.bit.IFMODE;
}

void MC3630::setWrapAddress9(bool enable)
{
  _registers.freg_2.bit.WRAPA = enable;
  writeReg(&_registers.freg_2, sizeof(_registers.freg_2));
}

void MC3630::setSniffSWakeRearm(bool enable)
{
  _registers.freg_1.bit.INTSC_EN = enable;
  writeReg(&_registers.freg_1, sizeof(_registers.freg_1));
}

void MC3630::setSpiStatus(bool enable)
{
  if (_registers.freg_1.bit.IFMODE & (IFMODE_SPI | IFMODE_SPI3))
  {
    _registers.freg_2.bit.SPI_STAT_EN = enable;
    writeReg(&_registers.freg_2, sizeof(_registers.freg_2));
  }
}

void MC3630::setI2CWrClear(bool enable)
{
  if (_registers.freg_1.bit.IFMODE & IFMODE_I2C)
  {
    _registers.freg_2.bit.I2CINT_WRCLRE = enable;
    writeReg(&_registers.freg_2, sizeof(_registers.freg_2));
  }
}

void MC3630::setExtTrigger(bool enable, bool polarity)
{
  _registers.freg_2.bit.EXT_TRIG_EN  = enable;
  _registers.freg_2.bit.EXT_TRIG_POL = polarity;
  writeReg(&_registers.freg_2, sizeof(_registers.freg_2));
}

void MC3630::setFifoStatus(bool enable)
{
  _registers.freg_2.bit.FIFO_STAT_EN = enable;
  writeReg(&_registers.freg_2, sizeof(_registers.freg_2));
}

void MC3630::setFifoMode(bool enable, uint8_t threshold, FIFO_OPT option)
{
  _registers.fifo.bit.TH             = (threshold < 31 ? threshold : 31);
  _registers.fifo.bit.MODE           = (option & FIFO_OPT_WATERMARK) != 0;
  _registers.fifo.bit.EN             = enable;
  writeReg(&_registers.fifo, sizeof(_registers.fifo));
  _registers.freg_1.bit.FREEZE       = (option & FIFO_OPT_FREEZE)    != 0;
  _registers.freg_2.bit.FIFO_BURST   = (option & FIFO_OPT_BURST)     != 0;
  _registers.freg_2.bit.FIFO_STREAM  = (option & FIFO_OPT_STREAM)    != 0;
  writeReg(&_registers.freg_1, sizeof(_registers.freg_1) + sizeof(_registers.freg_2));
}

void MC3630::resetFifo(void)
{
  _registers.fifo.bit.RESET = true;
  writeReg(&_registers.fifo, sizeof(_registers.fifo));
  _registers.fifo.bit.RESET = false;
}

void MC3630::disableAxis(uint16_t mask)
{
  _registers.mode.bit.X_AXIS_PD = (mask & (1 << AXIS_X)) != 0;
  _registers.mode.bit.Y_AXIS_PD = (mask & (1 << AXIS_Y)) != 0;
  _registers.mode.bit.Z_AXIS_PD = (mask & (1 << AXIS_Z)) != 0;
  writeReg(&_registers.mode, sizeof(_registers.mode));
}

void MC3630::setMode(MODE mode)
{
  if ((mode > MODE_STANDBY) && (_registers.mode.bit.MCTRL > MODE_STANDBY))
    setMode(MODE_STANDBY);
  if (mode == MODE_SNIFF)
    setSniffPowerMode(_spm_sniff);
  if (mode == MODE_TRIGGER)
    setSniffPowerMode(_spm_trigc);
  _registers.mode.bit.MCTRL = mode;
  _registers.mode.bit.TRIG_CMD = false;
  writeReg(&_registers.mode, sizeof(_registers.mode));
}

MC3630::MODE MC3630::getMode(void)
{
  readReg(&_registers.mode, sizeof(_registers.mode));
  return (MODE)_registers.mode.bit.MCTRL;
}

void MC3630::trigger(void)
{
  if (_registers.mode.bit.MCTRL == MODE_TRIGGER)
  {
    _registers.mode.bit.TRIG_CMD = true;
    writeReg(&_registers.mode, sizeof(_registers.mode));
  }
}

void MC3630::setTriggerConfig(STB_RATE rate, uint8_t count)
{
  _spm_trigc = rate >> 4;
  _registers.sniff.bit.STB_RATE = rate & 0x0F;
  writeReg(&_registers.sniff, sizeof(_registers.sniff));
  if (_registers.rate.bit.MAN_SEL != MAN_SEL_DISABLE)
    setManualRate(MAN_SEL_DISABLE, 0);
  _registers.trigc = count;
  writeReg(&_registers.trigc, sizeof(_registers.trigc));
}

void MC3630::setSniffPowerMode(uint8 spm)
{
  if (_registers.pmcr.bit.SPM != spm)
  {
    _registers.pmcr.bit.SPM = spm;
    writeReg(&_registers.pmcr, sizeof(_registers.pmcr));
  }
}

void MC3630::setSniffConfig(SNIFF_RATE rate, bool andMode, bool c2b)
{
  _spm_sniff = rate >> 4;
  _registers.sniff.bit.SR = rate & 0x0F;
  writeReg(&_registers.sniff, sizeof(_registers.sniff));
  _registers.sniffth.bit.AND_OR = andMode;
  _registers.sniffth.bit.MODE   = c2b;
  writeReg(&_registers.sniffth, sizeof(_registers.sniffth));
}

void MC3630::setSniffThreshold(THADR addr, uint8_t value, MUX_DELTA mux)
{
  if (addr >= THADR_COUNT_X)
  {
    if (value-- == 0)
      return;
    _registers.sniffcf.bit.CNTEN = true;
  }
  _registers.sniffcf.bit.THADR   = addr;
  _registers.sniffcf.bit.MUX     = mux;
  writeReg(&_registers.sniffcf, sizeof(_registers.sniffcf));
  _registers.sniffth.bit.TH      = (value < 63 ? value : 63);
  writeReg(&_registers.sniffth, sizeof(_registers.sniffth));
}

void MC3630::resetSniff(void)
{
  _registers.sniffcf.bit.RESET = true;
  writeReg(&_registers.sniffcf, sizeof(_registers.sniffcf));
  _registers.sniffcf.bit.RESET = false;
}

void MC3630::setWakeConfig(RATE rate, RANGE range, RES res)
{
  _registers.pmcr.bit.CSPM   = rate >> 4;
  writeReg(&_registers.pmcr, sizeof(_registers.pmcr));
  _registers.rate.bit.WR     = rate & 0x0F;
  writeReg(&_registers.rate, sizeof(_registers.rate));
  _registers.range.bit.RANGE = range;
  _registers.range.bit.RES   = res;
  writeReg(&_registers.range, sizeof(_registers.range));
}

void MC3630::setManualRate(MAN_SEL sel, uint8_t val)
{
  _registers.rate.bit.MAN_SEL = sel;
  writeReg(&_registers.rate, sizeof(_registers.rate));
  if (sel != MAN_SEL_DISABLE)
  {
    _registers.trigc = val;
    writeReg(&_registers.trigc, sizeof(_registers.trigc));
  }
}

void MC3630::setInterrupt(INTE enable, bool ipp, bool iah)
{
  _registers.intr.bit.IPP             = ipp;
  _registers.intr.bit.IAH             = iah;
  _registers.intr.bit.INT_WAKE        = (enable & INTE_WAKE)        != 0;
  _registers.intr.bit.INT_ACQ         = (enable & INTE_ACQ)         != 0;
  _registers.intr.bit.INT_FIFO_EMPTY  = (enable & INTE_FIFO_EMPTY)  != 0;
  _registers.intr.bit.INT_FIFO_FULL   = (enable & INTE_FIFO_FULL)   != 0;
  _registers.intr.bit.INT_FIFO_THRESH = (enable & INTE_FIFO_THRESH) != 0;
  _registers.intr.bit.INT_SWAKE       = (enable & INTE_SWAKE)       != 0;
  writeReg(&_registers.intr, sizeof(_registers.intr));
}

void MC3630::clearInterrupt(void)
{
  writeReg(&_registers.status_2, sizeof(_registers.status_2));
}

uint8_t MC3630::getChipID(void)
{
  readReg(&_registers.chip_id, sizeof(_registers.chip_id));
  return _registers.chip_id;
}

void MC3630::setScratch(uint8_t value)
{
  _registers.scratch = value;
  writeReg(&_registers.scratch, sizeof(_registers.scratch));
}

uint8_t MC3630::getScratch(void)
{
  readReg(&_registers.scratch, sizeof(_registers.scratch));
  return _registers.scratch;
}

void MC3630::setSpiHsMode(void)
{
  if (_registers.freg_1.bit.IFMODE & (IFMODE_SPI | IFMODE_SPI3))
  {
    _registers.pmcr.bit.SPI_HS_EN = true;
    writeReg(&_registers.pmcr, sizeof(_registers.pmcr));
    setSpiHsModeIO();
  }
}

void MC3630::reset(bool reset, bool reload)
{
  _registers.reset.bit.RESET  = reset;
  _registers.reset.bit.RELOAD = reload;
  writeReg(&_registers.reset, sizeof(_registers.reset));
}

void MC3630::init1(void)
{
  _registers.init_1 = 0x42;
  writeReg(&_registers.init_1, sizeof(_registers.init_1));
}

void MC3630::init2(void)
{
  _registers.init_2 = 0x00;
  writeReg(&_registers.init_2, sizeof(_registers.init_2));
}

void MC3630::init3(void)
{
  _registers.init_3 = 0x00;
  writeReg(&_registers.init_3, sizeof(_registers.init_3));
}

void MC3630::initDriveMotion(void)
{
  _registers.dm[AXIS_X].reg = 0x01;
  writeReg(&_registers.dm[AXIS_X], sizeof(_registers.dm[AXIS_X]));
  _registers.dm[AXIS_Y].reg = 0x80;
  writeReg(&_registers.dm[AXIS_Y], sizeof(_registers.dm[AXIS_Y]));
}

void MC3630::writeReg(const void *reg, uint8_t len)
{
  const uint8_t *buf = (const uint8_t *)reg;
  uint8_t  cmd = buf - (uint8_t *)&_registers;
  if (_registers.freg_1.bit.IFMODE & (IFMODE_SPI | IFMODE_SPI3))
    cmd |= 0x40;
  writeIO(cmd, buf, len);
}

void MC3630::readReg(void *reg, uint8_t len)
{
  uint8_t *buf = (uint8_t *)reg;
  uint8_t  cmd = buf - (uint8_t *)&_registers;
  if (_registers.freg_1.bit.IFMODE & (IFMODE_SPI | IFMODE_SPI3))
    cmd |= 0xC0;
  readIO(cmd, buf, len);
}

void MC3630::initialize(void)
{
  setMode(MODE_STANDBY);    // Go to Standby
  reset(true);              // Reset (or Power-On)
  delayMicroseconds(1000);  // Wait for reset to complete
  clearInterrupt();         // !! good luck charm !!
  delayMicroseconds(1000);  // Wait for state machine
  init1();                  // Initialization
  initDriveMotion();        // Initialization
  init2();                  // Initialization
  init3();                  // Initialization
}

void MC3630::writeIO(uint8_t cmd, const uint8_t *buf, uint8_t len)
{
  _iomethod->writeIO(cmd, buf, len);
}

void MC3630::readIO(uint8_t cmd, uint8_t *buf, uint8_t len)
{
  _iomethod->readIO(cmd, buf, len);
}

void MC3630::setSpiHsModeIO(void)
{
  _iomethod->setSpiHsModeIO();
}

void MC3630::delayMicroseconds(uint32 us)
{
  _iomethod->delayMicroseconds(us);
}
