/*
  mc3630.h - 3-Axis Accelerometer Sensor Library for mCube - MC3630

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include "stdint.h"
#include "stdbool.h"

#define MC3630_I2C_SLAVE_0 0x4C // DOUT_A1 = GND
#define MC3630_I2C_SLAVE_1 0x6C // DOUT_A1 = VDD

#define MC3630_NUMBER_OF_AXIS 3

#if defined(JENNIC_CHIP_FAMILY_JN516x)
  #undef  _BIT_FIELDS_HTOL
  #define _BIT_FIELDS_HTOL
  #define virtual  /* do not removed */
#endif

class MC3630
{
  private:

    typedef enum
    {                        // CWAKE, SWAKE Power Mode
      CSPM_LP_MODE  = 0b000, // Low Power Mode (nominal noise levels) (default)
      CSPM_ULP_MODE = 0b011, // Ultra-Low Power Mode (highest noise levels)
      CSPM_PR_MODE  = 0b100, // Precision Mode (lowest noise levels)
    } CSPM;

    typedef enum
    {                       // SNIFF Power Mode
      SPM_LP_MODE  = 0b000, // Low Power Mode (nominal noise levels) (default)
      SPM_ULP_MODE = 0b011, // Ultra-Low Power Mode (highest noise levels)
      SPM_PR_MODE  = 0b100, // Precision Mode (lowest noise levels)
    } SPM;

  public:

    typedef struct
    {
      void (*writeIO)(uint8_t cmd, const uint8_t *buf, uint8_t len);
      void (*readIO)(uint8_t cmd, uint8_t *buf, uint8_t len);
      void (*setSpiHsModeIO)(void);
      void (*delayMicroseconds)(uint32 us);
    } IOMETHOD;

    typedef enum
    {
      IFMODE_SPI3 = 1,
      IFMODE_I2C  = 2,
      IFMODE_SPI  = 4,
    } IFMODE;

    typedef enum
    {
      AXIS_X = 0,
      AXIS_Y = 1,
      AXIS_Z = 2,
    } AXIS;

    typedef enum
    {
      MODE_SLEEP   = 0b000,
      MODE_STANDBY = 0b001,
      MODE_SNIFF   = 0b010,
      MODE_CWAKE   = 0b101,
      MODE_SWAKE   = 0b110,
      MODE_TRIGGER = 0b111,
    } MODE;

    typedef enum
    {
      RES_6BIT  = 0b000, // 6 bits
      RES_7BIT  = 0b001, // 7 bits
      RES_8BIT  = 0b010, // 8 bits
      RES_10BIT = 0b011, // 10 bits
      RES_12BIT = 0b100, // 12 bits
      RES_14BIT = 0b101, // 14 bits (only 12-bits if FIFO enabled)
    } RES;

    typedef enum
    {
      RANGE_2G  = 0b000, // ±2g
      RANGE_4G  = 0b001, // ±4g
      RANGE_8G  = 0b010, // ±8g
      RANGE_16G = 0b011, // ±16g
      RANGE_12G = 0b100, // ±12g
    } RANGE;

    typedef enum
    {
      MOTION_DISABLE  = 0,
      MOTION_POSITIVE = 1,
      MOTION_NEGATIVE = 2,
      MOTION_BOTH     = 3,
    } MOTION;

    typedef enum
    {
      /* Ultra-Low Power */
      RATE_ULP_25Hz   = 0x06 | (CSPM_ULP_MODE << 4), //  0.9uA
      RATE_ULP_50Hz   = 0x07 | (CSPM_ULP_MODE << 4), //  1.6uA
      RATE_ULP_100Hz  = 0x08 | (CSPM_ULP_MODE << 4), //  2.8uA
      RATE_ULP_190Hz  = 0x09 | (CSPM_ULP_MODE << 4), //  5.5uA
      RATE_ULP_380Hz  = 0x0A | (CSPM_ULP_MODE << 4), // 10.0uA
      RATE_ULP_750Hz  = 0x0B | (CSPM_ULP_MODE << 4), // 18.0uA
      RATE_ULP_1100Hz = 0x0C | (CSPM_ULP_MODE << 4), // 26.0uA
      RATE_ULP_1300Hz = 0x0F | (CSPM_ULP_MODE << 4), // 36.0uA
      RATE_ULP_SLOW   = RATE_ULP_25Hz,
      RATE_ULP_FAST   = RATE_ULP_1300Hz,
      RATE_ULP_MANUAL = RATE_ULP_1300Hz,             // setManualRate()
      /* Low Power */
      RATE_LP_14Hz    = 0x05 | (CSPM_LP_MODE << 4), //  1.0uA
      RATE_LP_28Hz    = 0x06 | (CSPM_LP_MODE << 4), //  1.6uA
      RATE_LP_54Hz    = 0x07 | (CSPM_LP_MODE << 4), //  2.7uA
      RATE_LP_105Hz   = 0x08 | (CSPM_LP_MODE << 4), //  5.0uA
      RATE_LP_210Hz   = 0x09 | (CSPM_LP_MODE << 4), // 11.0uA
      RATE_LP_400Hz   = 0x0A | (CSPM_LP_MODE << 4), // 19.0uA
      RATE_LP_600Hz   = 0x0B | (CSPM_LP_MODE << 4), // 26.0uA
      RATE_LP_750Hz   = 0x0F | (CSPM_LP_MODE << 4), // 36.0uA
      RATE_LP_SLOW    = RATE_LP_14Hz,
      RATE_LP_FAST    = RATE_LP_750Hz,
      RATE_LP_MANUAL  = RATE_LP_750Hz,              // setManualRate()
      /* Precision */
      RATE_PR_14Hz    = 0x05 | (CSPM_PR_MODE << 4), //  5.0uA
      RATE_PR_28Hz    = 0x06 | (CSPM_PR_MODE << 4), // 10.0uA
      RATE_PR_55Hz    = 0x07 | (CSPM_PR_MODE << 4), // 18.0uA
      RATE_PR_80Hz    = 0x08 | (CSPM_PR_MODE << 4), // 25.0uA
      RATE_PR_100Hz   = 0x0F | (CSPM_PR_MODE << 4), // 36.0uA
      RATE_PR_SLOW    = RATE_PR_14Hz,
      RATE_PR_FAST    = RATE_PR_100Hz,
      RATE_PR_MANUAL  = RATE_PR_100Hz,              // setManualRate()
    } RATE;

    typedef enum
    {
      MAN_SEL_DISABLE          = 0b000, // Manual rate mode disabled. Normal operation
      MAN_SEL_RATE_IBTHC       = 0b001, // {SNIFF_RATE_IBTHC[3:0], WAKE_RATE_IBTHC[3:0]}
      MAN_SEL_WAKE_RATE_FPICK  = 0b010, // {0,0,0, WAKE_RATE_FPICK[4:0]}
      MAN_SEL_SNIFF_RATE_FPICK = 0b011, // {0,0,0, SNIFF_RATE_FPICK[4:0]}
      MAN_SEL_WAKE_RATE_LSB    = 0b100, // {WAKE_RATE_CLK_COUNT_LSB[7:0]}
      MAN_SEL_WAKE_RATE_MSB    = 0b101, // {WAKE_RATE_CLK_COUNT_MSB[15:8]}
      MAN_SEL_SNIFF_RATE_LSB   = 0b110, // {SNIFF_RATE_CLK_COUNT_LSB[7:0]}
      MAN_SEL_SNIFF_RATE_MSB   = 0b111, // {SNIFF_RATE_CLK_COUNT_MSB[15:8]}
    } MAN_SEL;

    typedef enum
    {
      /* Ultra-Low Power */
      SNIFF_RATE_ULP_DEFAULT = 0x00 | (SPM_ULP_MODE << 4), // 6Hz 0.4uA
      SNIFF_RATE_ULP_0p4Hz   = 0x01 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_0p8Hz   = 0x02 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_1p5Hz   = 0x03 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_6Hz     = 0x04 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_13Hz    = 0x05 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_25Hz    = 0x06 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_50Hz    = 0x07 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_100Hz   = 0x08 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_190Hz   = 0x09 | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_380Hz   = 0x0A | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_750Hz   = 0x0B | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_1100Hz  = 0x0C | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_1300Hz  = 0x0F | (SPM_ULP_MODE << 4),
      SNIFF_RATE_ULP_SLOW    = SNIFF_RATE_ULP_0p4Hz,
      SNIFF_RATE_ULP_FAST    = SNIFF_RATE_ULP_1300Hz,
      /* Low Power */
      SNIFF_RATE_LP_DEFAULT = 0x00 | (SPM_LP_MODE << 4), // 7Hz
      SNIFF_RATE_LP_0p4Hz   = 0x01 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_0p8Hz   = 0x02 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_1p5Hz   = 0x03 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_7Hz     = 0x04 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_14Hz    = 0x05 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_28Hz    = 0x06 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_54Hz    = 0x07 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_105Hz   = 0x08 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_210Hz   = 0x09 | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_400Hz   = 0x0A | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_600Hz   = 0x0B | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_750Hz   = 0x0F | (SPM_LP_MODE << 4),
      SNIFF_RATE_LP_SLOW    = SNIFF_RATE_LP_0p4Hz,
      SNIFF_RATE_LP_FAST    = SNIFF_RATE_LP_750Hz,
      /* Precision */
      SNIFF_RATE_PR_DEFAULT = 0x00 | (SPM_PR_MODE << 4), // 7Hz
      SNIFF_RATE_PR_0p2Hz   = 0x01 | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_0p4Hz   = 0x02 | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_0p9Hz   = 0x03 | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_7Hz     = 0x04 | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_14Hz    = 0x05 | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_28Hz    = 0x06 | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_55Hz    = 0x07 | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_80Hz    = 0x08 | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_100Hz   = 0x0F | (SPM_PR_MODE << 4),
      SNIFF_RATE_PR_SLOW    = SNIFF_RATE_PR_0p2Hz,
      SNIFF_RATE_PR_FAST    = SNIFF_RATE_PR_100Hz,
    } SNIFF_RATE;

    typedef enum
    {
      /* Ultra-Low Power */
      STB_RATE_ULP_1Hz   = 0x00 | (SPM_ULP_MODE << 4),
      STB_RATE_ULP_3Hz   = 0x01 | (SPM_ULP_MODE << 4),
      STB_RATE_ULP_5Hz   = 0x02 | (SPM_ULP_MODE << 4),
      STB_RATE_ULP_10Hz  = 0x03 | (SPM_ULP_MODE << 4),
      STB_RATE_ULP_23Hz  = 0x04 | (SPM_ULP_MODE << 4),
      STB_RATE_ULP_45Hz  = 0x05 | (SPM_ULP_MODE << 4),
      STB_RATE_ULP_90Hz  = 0x06 | (SPM_ULP_MODE << 4),
      STB_RATE_ULP_180Hz = 0x07 | (SPM_ULP_MODE << 4),
      STB_RATE_ULP_SLOW  = STB_RATE_ULP_1Hz,
      STB_RATE_ULP_FAST  = STB_RATE_ULP_180Hz,
      /* Low Power */
      STB_RATE_LP_0p5Hz  = 0x00 | (SPM_LP_MODE << 4),
      STB_RATE_LP_1Hz    = 0x01 | (SPM_LP_MODE << 4),
      STB_RATE_LP_3Hz    = 0x02 | (SPM_LP_MODE << 4),
      STB_RATE_LP_6Hz    = 0x03 | (SPM_LP_MODE << 4),
      STB_RATE_LP_12Hz   = 0x04 | (SPM_LP_MODE << 4),
      STB_RATE_LP_24Hz   = 0x05 | (SPM_LP_MODE << 4),
      STB_RATE_LP_48Hz   = 0x06 | (SPM_LP_MODE << 4),
      STB_RATE_LP_100Hz  = 0x07 | (SPM_LP_MODE << 4),
      STB_RATE_LP_SLOW   = STB_RATE_LP_0p5Hz,
      STB_RATE_LP_FAST   = STB_RATE_LP_100Hz,
      /* Precision */
      STB_RATE_PR_0p1Hz  = 0x00 | (SPM_PR_MODE << 4),
      STB_RATE_PR_0p2Hz  = 0x01 | (SPM_PR_MODE << 4),
      STB_RATE_PR_0p4Hz  = 0x02 | (SPM_PR_MODE << 4),
      STB_RATE_PR_0p8Hz  = 0x03 | (SPM_PR_MODE << 4),
      STB_RATE_PR_1p5Hz  = 0x04 | (SPM_PR_MODE << 4),
      STB_RATE_PR_3Hz    = 0x05 | (SPM_PR_MODE << 4),
      STB_RATE_PR_5Hz    = 0x06 | (SPM_PR_MODE << 4),
      STB_RATE_PR_10Hz   = 0x07 | (SPM_PR_MODE << 4),
      STB_RATE_PR_SLOW   = STB_RATE_PR_0p1Hz,
      STB_RATE_PR_FAST   = STB_RATE_PR_10Hz,
    } STB_RATE;

    typedef enum
    {
      THADR_NONE        = 0b000,
      THADR_THRESHOLD_X = 0b001,
      THADR_THRESHOLD_Y = 0b010,
      THADR_THRESHOLD_Z = 0b011,
      THADR_COUNT_X     = 0b101,
      THADR_COUNT_Y     = 0b110,
      THADR_COUNT_Z     = 0b111,
    } THADR;

    typedef enum
    {
      MUX_DELTA_5_0  = 0b000,
      MUX_DELTA_6_1  = 0b001,
      MUX_DELTA_7_2  = 0b010,
      MUX_DELTA_8_3  = 0b011,
      MUX_DELTA_9_4  = 0b100,
      MUX_DELTA_10_5 = 0b101,
    } MUX_DELTA;

    typedef enum
    {
      FIFO_OPT_WATERMARK = 0x01,
      FIFO_OPT_FREEZE    = 0x02,
      FIFO_OPT_BURST     = 0x04,
      FIFO_OPT_STREAM    = 0x08,
    } FIFO_OPT; // bit combination

    typedef enum
    {
      INTE_WAKE        = 0x01,
      INTE_ACQ         = 0x02,
      INTE_FIFO_EMPTY  = 0x04,
      INTE_FIFO_FULL   = 0x08,
      INTE_FIFO_THRESH = 0x10,
      INTE_SWAKE       = 0x20,
    } INTE; // bit combination

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t __RSV1__    : 4;
        uint8_t I2C_AD0_BIT : 1;
        uint8_t __RSV0__    : 3;
#else
        uint8_t __RSV0__    : 3;
        uint8_t I2C_AD0_BIT : 1;
        uint8_t __RSV1__    : 4;
#endif
      } bit;
      uint8_t reg;
    } EXT_STAT_1;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t SNIFF_DETECT: 1;
        uint8_t SNIFF_EN    : 1;
        uint8_t OTP_BUSY    : 1;
        uint8_t __RSV0__    : 3;
        uint8_t PD_CLK_STAT : 1;
        uint8_t OVR_DATA    : 1;
#else
        uint8_t OVR_DATA    : 1;
        uint8_t PD_CLK_STAT : 1;
        uint8_t __RSV0__    : 3;
        uint8_t OTP_BUSY    : 1;
        uint8_t SNIFF_EN    : 1;
        uint8_t SNIFF_DETECT: 1;
#endif
      } bit;
      uint8_t reg;
    } EXT_STAT_2;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t INT_PEND   : 1;
        uint8_t FIFO_THRESH: 1;
        uint8_t FIFO_FULL  : 1;
        uint8_t FIFO_EMPTY : 1;
        uint8_t NEW_DATA   : 1;
        uint8_t MODE       : 3;
#else
        uint8_t MODE       : 3;
        uint8_t NEW_DATA   : 1;
        uint8_t FIFO_EMPTY : 1;
        uint8_t FIFO_FULL  : 1;
        uint8_t FIFO_THRESH: 1;
        uint8_t INT_PEND   : 1;
#endif
      } bit;
      uint8_t reg;
    } STATUS_1;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t INT_SWAKE      : 1;
        uint8_t INT_FIFO_THRESH: 1;
        uint8_t INT_FIFO_FULL  : 1;
        uint8_t INT_FIFO_EMPTY : 1;
        uint8_t INT_ACQ        : 1;
        uint8_t INT_WAKE       : 1;
        uint8_t __RSV0__       : 2;
#else
        uint8_t __RSV0__       : 2;
        uint8_t INT_WAKE       : 1;
        uint8_t INT_ACQ        : 1;
        uint8_t INT_FIFO_EMPTY : 1;
        uint8_t INT_FIFO_FULL  : 1;
        uint8_t INT_FIFO_THRESH: 1;
        uint8_t INT_SWAKE      : 1;
#endif
      } bit;
      uint8_t reg;
    } STATUS_2;

    bool begin(IFMODE mode, IOMETHOD *iomethod = 0);
    EXT_STAT_1 getExtStatus1(void);
    EXT_STAT_2 getExtStatus2(void);
    STATUS_1 getStatus1(void);
    STATUS_2 getStatus2(void);
    void getOutput(int16_t axis[MC3630_NUMBER_OF_AXIS]);
    void setOffset(const int16_t axis[MC3630_NUMBER_OF_AXIS]);
    void setGain(const int16_t axis[MC3630_NUMBER_OF_AXIS]);
    void disableDriveMotion(const MOTION axis[MC3630_NUMBER_OF_AXIS]);
    void setSniffSWakeRearm(bool enable);
    void setSpiStatus(bool enable);
    void disableAxis(uint16_t mask);
    void setMode(MODE mode);
    MODE getMode(void);
    void trigger(void);
    void setTriggerConfig(STB_RATE rate, uint8_t count = 1);
    void setExtTrigger(bool enable, bool polarity);
    void setFifoStatus(bool enable);
    void setFifoMode(bool enable, uint8_t threshold, FIFO_OPT option);
    void resetFifo(void);
    void setSniffConfig(SNIFF_RATE rate, bool andMode = false, bool c2b = false);
    void setSniffThreshold(THADR addr, uint8_t value, MUX_DELTA mux = MUX_DELTA_5_0);
    void resetSniff(void);
    void setWakeConfig(RATE rate, RANGE range, RES res);
    void setManualRate(MAN_SEL sel, uint8_t val);
    void setInterrupt(INTE enable, bool ipp = false, bool iah = false);
    void clearInterrupt(void);
    uint8_t getChipID(void);
    void setScratch(uint8_t value);
    uint8_t getScratch(void);

  private:

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t IFMODE  : 3;
        uint8_t INTSC_EN: 1;
        uint8_t FREEZE  : 1;
        uint8_t ZERO    : 3;
#else
        uint8_t ZERO    : 3;
        uint8_t FREEZE  : 1;
        uint8_t INTSC_EN: 1;
        uint8_t IFMODE  : 3;
#endif
      } bit;
      uint8_t reg;
    } FREG_1;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t EXT_TRIG_EN  : 1;
        uint8_t EXT_TRIG_POL : 1;
        uint8_t FIFO_STREAM  : 1;
        uint8_t I2CINT_WRCLRE: 1;
        uint8_t FIFO_STAT_EN : 1;
        uint8_t SPI_STAT_EN  : 1;
        uint8_t FIFO_BURST   : 1;
        uint8_t WRAPA        : 1;
#else
        uint8_t WRAPA        : 1;
        uint8_t FIFO_BURST   : 1;
        uint8_t SPI_STAT_EN  : 1;
        uint8_t FIFO_STAT_EN : 1;
        uint8_t I2CINT_WRCLRE: 1;
        uint8_t FIFO_STREAM  : 1;
        uint8_t EXT_TRIG_POL : 1;
        uint8_t EXT_TRIG_EN  : 1;
#endif
      } bit;
      uint8_t reg;
    } FREG_2;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t TRIG_CMD : 1;
        uint8_t Z_AXIS_PD: 1;
        uint8_t Y_AXIS_PD: 1;
        uint8_t X_AXIS_PD: 1;
        uint8_t __RSV0__ : 1;
        uint8_t MCTRL    : 3;
#else
        uint8_t MCTRL    : 3;
        uint8_t __RSV0__ : 1;
        uint8_t X_AXIS_PD: 1;
        uint8_t Y_AXIS_PD: 1;
        uint8_t Z_AXIS_PD: 1;
        uint8_t TRIG_CMD : 1;
#endif
      } bit;
      uint8_t reg;
    } MODE_C;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t __RSV0__: 1;
        uint8_t MAN_SEL : 3;
        uint8_t WR      : 4;
#else
        uint8_t WR      : 4;
        uint8_t MAN_SEL : 3;
        uint8_t __RSV0__: 1;
#endif
      } bit;
      uint8_t reg;
    } RATE_1;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t STB_RATE: 3;
        uint8_t ZERO    : 1;
        uint8_t SR      : 4;
#else
        uint8_t SR      : 4;
        uint8_t ZERO    : 1;
        uint8_t STB_RATE: 3;
#endif
      } bit;
      uint8_t reg;
    } SNIFF_C;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t MODE  : 1;
        uint8_t AND_OR: 1;
        uint8_t TH    : 6;
#else
        uint8_t TH    : 6;
        uint8_t AND_OR: 1;
        uint8_t MODE  : 1;
#endif
      } bit;
      uint8_t reg;
    } SNIFFTH_C;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t RESET: 1;
        uint8_t MUX  : 3;
        uint8_t CNTEN: 1;
        uint8_t THADR: 3;
#else
        uint8_t THADR: 3;
        uint8_t CNTEN: 1;
        uint8_t MUX  : 3;
        uint8_t RESET: 1;
#endif
      } bit;
      uint8_t reg;
    } SNIFFCF_C;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t __RSV1__: 1;
        uint8_t RANGE   : 3;
        uint8_t __RSV0__: 1;
        uint8_t RES     : 3;
#else
        uint8_t RES     : 3;
        uint8_t __RSV0__: 1;
        uint8_t RANGE   : 3;
        uint8_t __RSV1__: 1;
#endif
      } bit;
      uint8_t reg;
    } RANGE_C;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t RESET: 1;
        uint8_t EN   : 1;
        uint8_t MODE : 1;
        uint8_t TH   : 5;
#else
        uint8_t TH   : 5;
        uint8_t MODE : 1;
        uint8_t EN   : 1;
        uint8_t RESET: 1;
#endif
      } bit;
      uint8_t reg;
    } FIFO_C;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t INT_SWAKE      : 1;
        uint8_t INT_FIFO_THRESH: 1;
        uint8_t INT_FIFO_FULL  : 1;
        uint8_t INT_FIFO_EMPTY : 1;
        uint8_t INT_ACQ        : 1;
        uint8_t INT_WAKE       : 1;
        uint8_t IAH            : 1;
        uint8_t IPP            : 1;
#else
        uint8_t IPP            : 1;
        uint8_t IAH            : 1;
        uint8_t INT_WAKE       : 1;
        uint8_t INT_ACQ        : 1;
        uint8_t INT_FIFO_EMPTY : 1;
        uint8_t INT_FIFO_FULL  : 1;
        uint8_t INT_FIFO_THRESH: 1;
        uint8_t INT_SWAKE      : 1;
#endif
      } bit;
      uint8_t reg;
    } INTR_C;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t SPI_HS_EN: 1;
        uint8_t SPM      : 3;
        uint8_t __RSV0__ : 1;
        uint8_t CSPM     : 3;
#else
        uint8_t CSPM     : 3;
        uint8_t __RSV0__ : 1;
        uint8_t SPM      : 3;
        uint8_t SPI_HS_EN: 1;
#endif
      } bit;
      uint8_t reg;
    } PMCR;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t __RSV1__: 4;
        uint8_t DN      : 1;
        uint8_t DP      : 1;
        uint8_t __RSV0__: 2;
#else
        uint8_t __RSV0__: 2;
        uint8_t DP      : 1;
        uint8_t DN      : 1;
        uint8_t __RSV1__: 4;
#endif
      } bit;
      uint8_t reg;
    } DM;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
#if defined(_BIT_FIELDS_HTOL)
        uint8_t RELOAD  : 1;
        uint8_t RESET   : 1;
        uint8_t __RSV0__: 6;
#else
        uint8_t __RSV0__: 6;
        uint8_t RESET   : 1;
        uint8_t RELOAD  : 1;
#endif
      } bit;
      uint8_t reg;
    } RESET;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
        uint8_t L;
#if defined(_BIT_FIELDS_HTOL)
        uint8_t GAIN8: 1;
        uint8_t H    : 7;
#else
        uint8_t H    : 7;
        uint8_t GAIN8: 1;
#endif
      } bit;
      struct __attribute__((packed))
      {
        uint8_t L;
        uint8_t H;
      } reg;
    } OFFSET;

    typedef union __attribute__((packed))
    {
      struct __attribute__((packed))
      {
        uint8_t L;
        uint8_t H;
      } reg;
    } OUTPUT;

    typedef struct __attribute__((packed))
    {
      EXT_STAT_1 ext_stat_1;                    // 0x00 R
      EXT_STAT_2 ext_stat_2;                    // 0x01 R
      OUTPUT     output[MC3630_NUMBER_OF_AXIS]; // 0x02 R
      STATUS_1   status_1;                      // 0x08 R
      STATUS_2   status_2;                      // 0x09 R
      uint8_t    __rsv0__[3];
      FREG_1     freg_1;                        // 0x0D W
      FREG_2     freg_2;                        // 0x0E W
      uint8_t    init_1;                        // 0x0F WO
      MODE_C     mode;                          // 0x10 W
      RATE_1     rate;                          // 0x11 W
      SNIFF_C    sniff;                         // 0x12 W
      SNIFFTH_C  sniffth;                       // 0x13 WS
      SNIFFCF_C  sniffcf;                       // 0x14 W
      RANGE_C    range;                         // 0x15 W
      FIFO_C     fifo;                          // 0x16 W
      INTR_C     intr;                          // 0x17 R
      uint8_t    chip_id;                       // 0x18 R
      uint8_t    __rsv1__;
      uint8_t    init_3;                        // 0x1A RW
      uint8_t    scratch;                       // 0x1B RW
      PMCR       pmcr;                          // 0x1C W
      uint8_t    __rsv2__[3];
      DM         dm[3];                         // 0x20 W
      uint8_t    __rsv3__;
      RESET      reset;                         // 0x24 W
      uint8_t    __rsv4__[3];
      uint8_t    init_2;                        // 0x28 W
      uint8_t    trigc;                         // 0x29 W
      OFFSET     offset[MC3630_NUMBER_OF_AXIS]; // 0x2A W
      uint8_t    gain[MC3630_NUMBER_OF_AXIS];   // 0x30 W
    } REGISTERS;

    IOMETHOD  *_iomethod;
    bool       _enable;
    REGISTERS  _registers;
    uint8      _spm_sniff;
    uint8      _spm_trigc;

    void setInterface(IFMODE mode);
    IFMODE getInterface(void);
    void setWrapAddress9(bool enable);
    void setI2CWrClear(bool enable);
    void setSpiHsMode(void);
    void reset(bool reset, bool reload = false);
    void setSniffPowerMode(uint8 spm);
    void initDriveMotion(void);
    void init1(void);
    void init2(void);
    void init3(void);
    void writeReg(const void *reg, uint8_t len);
    void readReg(void *reg, uint8_t len);
    void initialize(void);

  protected:

    virtual void writeIO(uint8_t cmd, const uint8_t *buf, uint8_t len);
    virtual void readIO(uint8_t cmd, uint8_t *buf, uint8_t len);
    virtual void setSpiHsModeIO(void);
    virtual void delayMicroseconds(uint32 us);
};
