/*
  spi.cpp - SPI Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "spi.h"

uint8 SpiConfig::_configured;

uint8 SSpi::_u8TxBuffer[];
uint8 SSpi::_u8RxBuffer[];

SpiConfig& SpiConfig::speed(SPEED value)
{
  if (_eSpeed != value)
    _changed = true;
  _eSpeed = value;
  return *this;
}

/*
   Enable/disable data transfer with the least significant bit (LSB) transferred first:
     TRUE  - enable
     FALSE - disable
*/
SpiConfig& SpiConfig::lsbFirst(bool value)
{
  if (_bLsbFirst != value)
    _changed = true;
  _bLsbFirst = value;
  return *this;
}

/*
   Number of SPI slaves to control. Valid values are 0-3 (higher values are truncated to 3)
 */
SpiConfig& SpiConfig::slaveEnable(uint8 value)
{
  if (_u8SlaveEnable != value)
    _changed = true;
  _u8SlaveEnable = value;
  return *this;
}

/*
   Enable/disable automatic slave selection:
     TRUE  - enable
     FALSE - disable
 */
SpiConfig& SpiConfig::autoSelect(bool value)
{
  if (_bAutoSelect != value)
    _changed = true;
  _bAutoSelect = value;
  return *this;
}

/*
   DIO on which specified slave select line will operate:
     TRUE  - DIO14 (SPISEL1) or DIO15 (SPISEL2)
     FALSE - DIO0  (SPISEL1) or DIO1  (SPISEL2)
*/
SpiConfig& SpiConfig::location(SELLOC sel, bool value)
{
  switch (sel)
  {
    case SELLOC_1:
      if (value)
        _bLocation |=  (1 << SELLOC_1);
      else
        _bLocation &= ~(1 << SELLOC_1);
      break;
    case SELLOC_2:
      if (value)
        _bLocation |=  (1 << SELLOC_2);
      else
        _bLocation &= ~(1 << SELLOC_2);
      break;
  }
  return *this;
}

SpiConfig& SpiConfig::delayReadEdge(bool value)
{
  _delayReadEdge = value;
  return *this;
}

/*
 * Pointer to callback function to be registered
 */
SpiConfig& SpiConfig::callback(PR_HWINT_APPCALLBACK prSpiCallback)
{
  if (_prSpiCallback != prSpiCallback)
    _changed = true;
  _prSpiCallback = prSpiCallback;
  return *this;
}

void SpiConfig::beginTransaction(void)
{
  if (!_changed)
    vAHI_SpiRestoreConfiguration(&_configuration);
  else
  {
    _changed = false;
    vAHI_SpiConfigure
    (
      _u8SlaveEnable,
      _bLsbFirst,
      _eMode >> 1,
      _eMode &  1,
      _eSpeed,
      _prSpiCallback != 0,
      _bAutoSelect
    );
    vAHI_SpiReadConfiguration(&_configuration);
  }
  vAHI_SpiSelSetLocation(E_AHI_SPISEL_1, _bLocation & (1 << SELLOC_1));
  vAHI_SpiSelSetLocation(E_AHI_SPISEL_2, _bLocation & (1 << SELLOC_2));
  vAHI_SpiSetDelayReadEdge(_delayReadEdge);
  if (_prSpiCallback)
    vAHI_SpiRegisterCallback(_prSpiCallback);
  _configured = true;
}

void SpiConfig::endTransaction(void)
{
}

void SpiConfig::end(void)
{
  if (_configured)
  {
    _configured = false;
    vAHI_SpiDisable();
  }
}

void Spi::select(SEL u8SlaveMask)
{
  vAHI_SpiSelect(u8SlaveMask);
}

void Spi::unselect(void)
{
  vAHI_SpiStop();
}

void Spi::write(uint16 slave, uint8 u8Cmd, const uint8 *pu8Write, uint16 u16WriteLen)
{
  select((SEL)slave);
  write(u8Cmd);
  write(pu8Write, u16WriteLen);
  unselect();
}

void Spi::writeAndRead(uint16 slave, uint8 u8Cmd, const uint8 *pu8Write, uint16 u16WriteLen, uint8 *pu8Read, uint16 u16ReadLen)
{
  select((SEL)slave);
  write(u8Cmd);
  write(pu8Write, u16WriteLen);
  read(pu8Read, u16ReadLen);
  unselect();
}

void Spi::writeAndRead(uint16 slave, const uint8 *pu8Write, uint16 u16WriteLen, uint8 *pu8Read, uint16 u16ReadLen)
{
  select((SEL)slave);
  write(pu8Write, u16WriteLen);
  read(pu8Read, u16ReadLen);
  unselect();
}

void Spi::write(uint8 u8Out)
{
  transfer(u8Out);
}

void Spi::write(const uint8 *pu8Buf, uint16 u16Len)
{
  if (pu8Buf)
  {
    while (u16Len--)
      write(*pu8Buf++);
  }
}

void Spi::write9(uint16 u16Out)
{
  transfer9(u16Out);
}

void Spi::write9(const uint16 *pu16Buf, uint16 u16Len)
{
  if (pu16Buf)
  {
    while (u16Len--)
      write9(*pu16Buf++);
  }
}

void Spi::write16(uint16 u16Out)
{
  transfer16(u16Out);
}

void Spi::write16(const uint16 *pu16Buf, uint16 u16Len)
{
  if (pu16Buf)
  {
    while (u16Len--)
      write16(*pu16Buf++);
  }
}

void Spi::write32(uint32 u32Out)
{
  transfer32(u32Out);
}

void Spi::write32(const uint32 *pu32Buf, uint16 u16Len)
{
  if (pu32Buf)
  {
    while (u16Len--)
      write32(*pu32Buf++);
  }
}

uint8 Spi::read(void)
{
  return transfer((uint8)-1);
}

void Spi::read(uint8 *pu8Buf, uint16 u16Len)
{
  if (pu8Buf)
  {
    for (uint32 i = 0; i < u16Len; ++i)
      *pu8Buf++ = read();
  }
}

uint16 Spi::read9(void)
{
  return transfer9((uint16)-1);
}

void Spi::read9(uint16 *pu16Buf, uint16 u16Len)
{
  if (pu16Buf)
  {
    for (uint32 i = 0; i < u16Len; ++i)
      *pu16Buf++ = read9();
  }
}

uint16 Spi::read16(void)
{
  return transfer16((uint16)-1);
}

void Spi::read16(uint16 *pu16Buf, uint16 u16Len)
{
  if (pu16Buf)
  {
    for (uint32 i = 0; i < u16Len; ++i)
      *pu16Buf++ = read16();
  }
}

uint32 Spi::read32(void)
{
  return transfer32((uint32)-1);
}

void Spi::read32(uint32 *pu32Buf, uint16 u16Len)
{
  if (pu32Buf)
  {
    for (uint32 i = 0; i < u16Len; ++i)
      *pu32Buf++ = read32();
  }
}

uint8 Spi::transfer(uint8 u8Out)
{
  return transfer(u8Out, 7);
}

uint16 Spi::transfer9(uint16 u16Out)
{
  return transfer(u16Out, 8) & 0x01FF;
}

uint16 Spi::transfer16(uint16 u16Out)
{
  return transfer(u16Out, 15);
}

uint32 Spi::transfer32(uint32 u32Out)
{
  return transfer(u32Out, 31);
}

uint32 Spi::transfer(uint32 u32Out, uint8 u8CharLen)
{
  vAHI_SpiStartTransfer(u8CharLen, u32Out);
  wait();
  return u32AHI_SpiReadTransfer32();
}

bool Spi::busy(void)
{
  // This function polls the SPI master to determine whether it is currently busy
  // performing a data transfer.
  return bAHI_SpiPollBusy();
}

void Spi::wait(void)
{
  // This function waits for the SPI master to complete a transfer and then returns.
  vAHI_SpiWaitBusy();
}

bool SSpi::begin(bool bPinLocation, bool bLsbFirst, /*INTMASK*/uint16 imask, PR_HWINT_APPCALLBACK prSpiCallback)
{
  if (prSpiCallback)
    vAHI_SpiSlaveRegisterCallback(prSpiCallback);
  else
    imask = 0;
  return bAHI_SpiSlaveEnable
  (
    bPinLocation,
    bLsbFirst,
    _u8TxBuffer,
    SPI_SLAVE_BUFFER_SIZE,
    SPI_SLAVE_BUFFER_SIZE * SPI_SLAVE_THRESHOLD / 100, // an interrupt will be generated when the amount of data in the buffer falls below this level
    _u8RxBuffer,
    SPI_SLAVE_BUFFER_SIZE,
    SPI_SLAVE_BUFFER_SIZE * (100 - SPI_SLAVE_THRESHOLD) / 100, // an interrupt will be generated when the amount of data in the buffer rises above this level
    SPI_SLAVE_TIMEOUT,
    imask
  );
}

void SSpi::end(void)
{
  vAHI_SpiSlaveDisable();
}

void SSpi::reset(bool bTxReset, bool bRxReset)
{
  vAHI_SpiSlaveReset(bTxReset, bRxReset);
}

uint8 SSpi::availableForWrite(void)
{
  return SPI_SLAVE_BUFFER_SIZE - u8AHI_SpiSlaveTxFillLevel();
}

uint16 SSpi::write8(uint8 u8Byte)
{
  if (!availableForWrite())
    return 0;
  vAHI_SpiSlaveTxWriteByte(u8Byte);
  return 1;
}

uint16 SSpi::write8(const uint8 *pu8Buf, uint16 u16Len)
{
  uint16 cnt;
  for (cnt = 0; cnt < u16Len; ++cnt)
  {
    if (!write8(*pu8Buf++))
      break;
  }
  return cnt;
}

uint8 SSpi::available(void)
{
  return u8AHI_SpiSlaveRxFillLevel();
}

int SSpi::read8(void)
{
  return available() ? u8AHI_SpiSlaveRxReadByte() : -1;
}

uint16 SSpi::read8(uint8 *pu8Buf, uint16 u16Len)
{
  uint16 cnt;
  for (cnt = 0; cnt < u16Len; ++cnt)
  {
    int b;
    if ((b = read8()) < 0)
      break;
    *pu8Buf++ = b;
  }
  return cnt;
}

uint8 SSpi::status(void)
{
  return u8AHI_SpiSlaveStatus();
}
