/*
  spi.h - SPI Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <jendefs.h>
#include <AppHardwareApi.h>

/*
   I/O Ports

     SCLK: DO0
     MISO: DO1
     MOSI: DIO18
     SEL0: DIO19
     SEL1: DIO0/DIO14
     SEL2: DIO1/DIO15
*/

#ifndef SPI_DEFAULT_SPEED
#define SPI_DEFAULT_SPEED SPEED_8MHZ
#endif

#ifndef SPI_SLAVE_BUFFER_SIZE
#define SPI_SLAVE_BUFFER_SIZE 128 // 1 to 255 bytes
#endif

#ifndef SPI_SLAVE_TIMEOUT
#define SPI_SLAVE_TIMEOUT 1000    // 0 to 4095 us
#endif

#ifndef SPI_SLAVE_THRESHOLD
#define SPI_SLAVE_THRESHOLD 10    // 0 to 100 %
#endif

class SpiConfig
{
  public:

    typedef enum
    {
      MODE_0 = 0,
      MODE_1 = 1,
      MODE_2 = 2,
      MODE_3 = 3,
    } MODE;

    typedef enum
    {
      SPEED_16MHZ = 0,
      SPEED_8MHZ  = 1,
      SPEED_4MHZ  = 2,
      SPEED_2MHZ  = 3,
      SPEED_1MHZ  = 4,
    } SPEED;

    /* must be first call */
    SpiConfig& mode(MODE mode)
    {
      _eMode            = mode;
      _eSpeed           = SPI_DEFAULT_SPEED;
      _bLsbFirst        = false;
      _u8SlaveEnable    = 2;     // bit0, 1, 2
      _bAutoSelect      = false;
      _delayReadEdge    = false;
      _prSpiCallback    = 0;
      _configuration    = 0;
      _bLocation        = 0;
      _changed          = true;
      return *this;
    }

    /*
       Clock divisor in the range 0 to 63. Peripheral clock is divided by 2 x u8ClockDivider,
       but 0 is a special value used when no clock division is required
     */
    SpiConfig& speed(SPEED value);

    /*
       Enable/disable data transfer with the least significant bit (LSB) transferred first:
         TRUE  - enable
         FALSE - disable
    */
    SpiConfig& lsbFirst(bool value);

    /*
       Number of SPI slaves to control. Valid values are 0-3 (higher values are truncated to 3)
     */
    SpiConfig& slaveEnable(uint8 value);

    /*
       Enable/disable automatic slave selection:
         TRUE  - enable
         FALSE - disable
     */
    SpiConfig& autoSelect(bool value);

    typedef enum
    {
      SELLOC_1 = E_AHI_SPISEL_1,
      SELLOC_2 = E_AHI_SPISEL_2,
    } SELLOC;

    /*
       DIO on which specified slave select line will operate:
         TRUE  - DIO14 (SPISEL1) or DIO15 (SPISEL2)
         FALSE - DIO0  (SPISEL1) or DIO1  (SPISEL2)
    */
    SpiConfig& location(SELLOC sel, bool value);

    SpiConfig& delayReadEdge(bool value);

    /*
     * Pointer to callback function to be registered
     */
    SpiConfig& callback(PR_HWINT_APPCALLBACK prSpiCallback);

    void beginTransaction(void);
    void endTransaction(void);

    static void end(void);

  private:

    tSpiConfiguration    _configuration;
    uint8                _bLocation;
    MODE                 _eMode;
    SPEED                _eSpeed;
    bool                 _bLsbFirst;
    uint8                _u8SlaveEnable;
    bool                 _bAutoSelect;
    bool                 _delayReadEdge;
    PR_HWINT_APPCALLBACK _prSpiCallback;
    uint8                _changed;

    static uint8         _configured;

    friend class Spi;
};

class Spi
{
  public:

    typedef enum
    {
      SEL_0 = 1, // DIO19
      SEL_1 = 2, // DIO0/DIO14
      SEL_2 = 4, // DIO1/DIO15
    } SEL;

    static void select(SEL u8SlaveMask);
    static void unselect(void);
    static void write(uint16 slave, uint8 u8Cmd, const uint8 *pu8Write, uint16 u16WriteLen);
    static void writeAndRead(uint16 slave, uint8 u8Cmd, const uint8 *pu8Write, uint16 u16WriteLen, uint8 *pu8Read, uint16 u16ReadLen);
    static void writeAndRead(uint16 slave, const uint8 *pu8Write, uint16 u16WriteLen, uint8 *pu8Read, uint16 u16ReadLen);
    static void write(uint8 u8Out);
    static void write(const uint8 *pu8Buf, uint16 u16Len);
    static void write9(uint16 u16Out);
    static void write9(const uint16 *pu16Buf, uint16 u16Len);
    static void write16(uint16 u16Out);
    static void write16(const uint16 *pu16Buf, uint16 u16Len);
    static void write32(uint32 u32Out);
    static void write32(const uint32 *pu32Buf, uint16 u16Len);
    static uint8 read(void);
    static void read(uint8 *pu8Buf, uint16 u16Len);
    static uint16 read9(void);
    static void read9(uint16 *pu16Buf, uint16 u16Len);
    static uint16 read16(void);
    static void read16(uint16 *pu16Buf, uint16 u16Len);
    static uint32 read32(void);
    static void read32(uint32 *pu32Buf, uint16 u16Len);
    static uint8 transfer(uint8 u8Out);
    static uint16 transfer9(uint16 u16Out);
    static uint16 transfer16(uint16 u16Out);
    static uint32 transfer32(uint32 u32Out);
    static uint32 transfer(uint32 u32Out, uint8 u8CharLen);

  private:

    static bool busy(void);
    static void wait(void);
};

class SSpi
{
  public:

    typedef enum
    {
      INTMASK_RX_FIRST   = E_AHI_SPIS_INT_RX_FIRST_MASK,
      INTMASK_TX_LAST    = E_AHI_SPIS_INT_TX_LAST_MASK,
      INTMASK_RX_CLIMB   = E_AHI_SPIS_INT_RX_CLIMB_MASK,
      INTMASK_TX_FALL    = E_AHI_SPIS_INT_TX_FALL_MASK,
      INTMASK_RX_OVER    = E_AHI_SPIS_INT_RX_OVER_MASK,
      INTMASK_TX_OVER    = E_AHI_SPIS_INT_TX_OVER_MASK,
      INTMASK_RX_UNDER   = E_AHI_SPIS_INT_RX_UNDER_MASK,
      INTMASK_TX_UNDER   = E_AHI_SPIS_INT_TX_UNDER_MASK,
      INTMASK_RX_TIMEOUT = E_AHI_SPIS_INT_RX_TIMEOUT_MASK,
    } INTMASK;

    static bool begin(bool bPinLocation = false, bool bLsbFirst = false, /*INTMASK*/uint16 imask = 0, PR_HWINT_APPCALLBACK prSpiCallback = 0);
    static void end(void);
    static void reset(bool bTxReset, bool bRxReset);
    static uint8 availableForWrite(void);
    static uint16 write8(uint8 u8Byte);
    static uint16 write8(const uint8 *pu8Buf, uint16 u16Len);
    static uint8 available(void);
    static int read8(void);
    static uint16 read8(uint8 *pu8Buf, uint16 u16Len);
    static uint8 status(void);

  private:

    static uint8 _u8TxBuffer[SPI_SLAVE_BUFFER_SIZE];
    static uint8 _u8RxBuffer[SPI_SLAVE_BUFFER_SIZE];
};
