/*
  system.cpp - System Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "device.h"

uint32 System::millis(void)
{
  return MulDiv::divu64u64(System::micros64(), 1000);
}

uint32 System::micros(void)
{
  return (uint32)TickTimer::micros64();
}

uint64 System::micros64(void)
{
  return TickTimer::micros64();
}

void System::delayMicroseconds(uint32 us, uint8 busyLoop)
{
  uint32 t = micros();
  if (us > busyLoop)
  {
    us -= busyLoop;
    while (micros() - t < us)
      yield();
    t += us;
  }
  if (busyLoop)
  {
    while (micros() - t < busyLoop)
      continue;
  }
}

void System::delayMicroseconds64(uint64 us, uint8 busyLoop)
{
  uint64 t = micros64();
  if (us > busyLoop)
  {
    us -= busyLoop;
    while (micros64() - t < us)
      yield();
    t += us;
  }
  if (busyLoop)
  {
    while (micros64() - t < busyLoop)
      continue;
  }
}

void System::delay(uint32 ms, uint8 busyLoop)
{
  if (ms <= ((uint32)-1 / 1000U))
    delayMicroseconds(ms * 1000UL, busyLoop);
  else
    delayMicroseconds64(MulDiv::mulu32u32(ms, 1000U), busyLoop);
}

uint32 System::firstBit(uint32 n)
{
  asm volatile ("b.ff1 %0, %1" : "=r" (n) : "r" (n));
  return n;
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wreturn-type"

uint32 System::firstBit64(uint64 n)
{
  asm volatile ("b.mov  r5, r3");
  asm volatile ("b.ff1  r3, r4");
  asm volatile ("b.bne  r3, r0, L_firstBit64_1");
  asm volatile ("b.ff1  r3, r5");
  asm volatile ("b.beq  r3, r0, L_firstBit64_1");
  asm volatile ("b.addi r3, r3, 32");
  asm volatile ("L_firstBit64_1:");
}

uint32 System::lastBit(uint32 n)
{
  asm volatile ("b.bitrev r3, r3");
  asm volatile ("b.ff1    r3, r3");
  asm volatile ("b.beq    r3, r0, L_lastBit_1");
  asm volatile ("b.addi   r5, r0, 33");
  asm volatile ("b.sub    r3, r5, r3");
  asm volatile ("L_lastBit_1:");
}

uint32 System::lastBit64(uint64 n)
{
  asm volatile ("b.addi   r5, r0, 65");
  asm volatile ("b.bitrev r3, r3");
  asm volatile ("b.ff1    r3, r3");
  asm volatile ("b.bne    r3, r0, L_lastBit64_1");
  asm volatile ("b.bitrev r3, r4");
  asm volatile ("b.ff1    r3, r3");
  asm volatile ("b.beq    r3, r0, L_lastBit64_2");
  asm volatile ("b.addi   r5, r0, 33");
  asm volatile ("L_lastBit64_1:");
  asm volatile ("b.sub    r3, r5, r3");
  asm volatile ("L_lastBit64_2:");
}

#pragma GCC diagnostic pop

void System::enableInterrupts(void)
{
  MICRO_ENABLE_INTERRUPTS();
}

void System::disableInterrupts(void)
{
  MICRO_DISABLE_INTERRUPTS();
}

uint32 System::disableAndSaveInterrupts(void)
{
  uint32 save;
  MICRO_DISABLE_AND_SAVE_INTERRUPTS(save);
  return save;
}

void System::restoreInterrupts(uint32 save)
{
  MICRO_RESTORE_INTERRUPTS(save);
}

void System::interruptPriority(uint16 u16Mask, uint8 u8Level)
{
  vAHI_InterruptSetPriority(u16Mask, u8Level);
}

void System::cpuDoze(void)
{
  vAHI_CpuDoze();
}

void System::sleep(SLEEP mode)
{
  vAHI_Sleep((teAHI_SleepMode)mode);
}

void System::protocolPower(bool enable)
{
  vAHI_ProtocolPower(enable);
}

System::POWER System::powerStatus(void)
{
  return (POWER)u16AHI_PowerStatus();
}

void System::reset(void)
{
  vAHI_SwReset();
}

void System::writeNVData(uint8 u8Location, uint32 u32WriteData)
{
  vAHI_WriteNVData(u8Location, u32WriteData);
}

uint32 System::readNVData(uint8 u8Location)
{
  return u32AHI_ReadNVData(u8Location);
}

void System::setStackOverflow(bool_t bStkOvfEn, uint32 u32Addr)
{
  vAHI_SetStackOverflow(bStkOvfEn, u32Addr);
}

void System::debugger(bool enable, bool location)
{
  vAHI_SetJTAGdebugger(enable, location);
}

void System::storeDebug(uint32 *pRegStorage)
{
  vAHI_StoreDebug(pRegStorage);
}

void System::restoreDebug(uint32 *pRegStorage)
{
  vAHI_RestoreDebug(pRegStorage);
}

/* JN516x Main Routine */

#if defined __cplusplus
#define APPSTART extern "C" PUBLIC
#else
#define APPSTART PUBLIC
#endif

extern void wakeup(void) __attribute__((weak));
extern void setup(void);
extern void loop(void);

extern void empty(uint32 u32Device, uint32 u32ItemBitmap) __attribute__((noinline,optimize(1)));
void empty(uint32 u32Device, uint32 u32ItemBitmap)
{
  (void)u32Device;
  (void)u32ItemBitmap;
  /* No processing. for wakeup　*/
}

void polling(void)
{
  Led::handle();
  Board::handle();
  MAdc::handle();
  Watchdog::reset();
  WPan::handle();
}

void yield(void)
{
  if (Task::yield())
    polling();
}

static
void init(void)
{
  u32AHI_Init();
  u32AppQApiInit(0, 0, 0);
  Gpio::begin();
  SysCtrl::begin();
  Watchdog::begin();
  WakeTimer::begin();
  TickTimer::begin();
  Analog::begin();
  EEPRom::begin();
  AlarmTimer.begin();
  Task::begin();
  Led::begin();
  Board::begin();
}

APPSTART
void AppColdStart(void)
{
  init();
  setup();
  while (1)
  {
    loop();
    yield();
  }
}

APPSTART
void AppWarmStart(void)
{
  if (wakeup)
    wakeup();
  AppColdStart();
}
