/*
  task.cpp - Smallest Non Preemptive Multi Task Library for Embedded System

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <alloca.h>
#include "system.h"
#include "task.h"

Task::tcb_t *Task::_disp;
Task::tcb_t *Task::_curr;
Task::tcb_t *Task::_head;
Task::tcb_t *Task::_tail[];
size_t       Task::_used;

#if defined(JENNIC_CHIP_FAMILY_JN516x)

int Task::setjmp(jn516x_jmp_buf __jmpb)
{
  asm volatile ("b.sw 0x00(r3), r1");  /* SP (Stack pointer) */
  asm volatile ("b.sw 0x04(r3), r2");
  asm volatile ("b.sw 0x08(r3), r9");  /* LR (Link address register) */
  asm volatile ("b.sw 0x0C(r3), r10"); /* FP (Frame pointer) */
  asm volatile ("b.sw 0x10(r3), r11");
  asm volatile ("b.sw 0x14(r3), r12");
  asm volatile ("b.sw 0x18(r3), r13");
  asm volatile ("b.sw 0x1C(r3), r14");
  asm volatile ("b.sw 0x20(r3), r15");
  asm volatile ("b.sw 0x24(r3), r16");
  asm volatile ("b.sw 0x28(r3), r17");
  asm volatile ("b.sw 0x2C(r3), r18");
  asm volatile ("b.sw 0x30(r3), r19");
  asm volatile ("b.sw 0x34(r3), r20");
  asm volatile ("b.sw 0x38(r3), r21");
  asm volatile ("b.sw 0x3C(r3), r22");
  asm volatile ("b.sw 0x40(r3), r23");
  asm volatile ("b.sw 0x44(r3), r24");
  asm volatile ("b.sw 0x48(r3), r25");
  asm volatile ("b.sw 0x4C(r3), r26");
  asm volatile ("b.sw 0x50(r3), r27");
  asm volatile ("b.sw 0x54(r3), r28");
  asm volatile ("b.sw 0x58(r3), r29");
  asm volatile ("b.sw 0x5C(r3), r30");
  asm volatile ("b.sw 0x60(r3), r31");
  return 0;
}

void Task::longjmp(jn516x_jmp_buf __jmpb, int __retval)
{
  asm volatile ("b.lwz  r31, 0x60(r3)");
  asm volatile ("b.lwz  r30, 0x5C(r3)");
  asm volatile ("b.lwz  r29, 0x58(r3)");
  asm volatile ("b.lwz  r28, 0x54(r3)");
  asm volatile ("b.lwz  r27, 0x50(r3)");
  asm volatile ("b.lwz  r26, 0x4C(r3)");
  asm volatile ("b.lwz  r25, 0x48(r3)");
  asm volatile ("b.lwz  r24, 0x44(r3)");
  asm volatile ("b.lwz  r23, 0x40(r3)");
  asm volatile ("b.lwz  r22, 0x3C(r3)");
  asm volatile ("b.lwz  r21, 0x38(r3)");
  asm volatile ("b.lwz  r20, 0x34(r3)");
//asm volatile ("b.lwz  r19, 0x30(r3)"); /* CPU-BUG? */
  asm volatile ("b.lwz  r18, 0x2C(r3)");
//asm volatile ("b.lwz  r17, 0x28(r3)"); /* CPU-BUG? */
  asm volatile ("b.lwz  r16, 0x24(r3)");
  asm volatile ("b.lwz  r15, 0x20(r3)");
  asm volatile ("b.lwz  r14, 0x1C(r3)");
  asm volatile ("b.lwz  r13, 0x18(r3)");
  asm volatile ("b.lwz  r12, 0x14(r3)");
  asm volatile ("b.lwz  r11, 0x10(r3)");
  asm volatile ("b.lwz  r10, 0x0C(r3)"); /* FP (Frame pointer) */
  asm volatile ("b.lwz  r9,  0x08(r3)"); /* LR (Link address register) */
  asm volatile ("b.lwz  r2,  0x04(r3)");
  asm volatile ("b.lwz  r1,  0x00(r3)"); /* SP (Stack pointer) */
  asm volatile ("b.mov  r3,  r4");
}

#endif

void Task::begin(void)
{
  /* Registration of main task */
  static tcb_t tcb;
  _curr = &tcb;
}

bool Task::start(func_t func, void *arg, int size, void *stack)
{
  tcb_t *tcb;
  if (size < 0)
  {
    /* Use the current stack area */
    size  = (_used += - size);
    stack = 0;
  }
  else if (stack)
  {
    /* Set the stack area to the variable area */
    size  = (char *)&tcb - (char *)stack - size + sizeof(*tcb);
    stack = 0;
  }
  else
  {
    /* Dynamically allocate stack area */
    stack = malloc(size);
    if (!stack)
      return false;
    size = (char *)&tcb - (char *)stack - size + sizeof(*tcb);
  }
  tcb = (tcb_t *)alloca(size);
  tcb->alloc = stack;
  tcb->arg   = arg;
  tcb->start = func;
  enqueue0(tcb);
  if (setjmp(tcb->context))
  {
    _curr = _disp;
    _curr->start(_curr->arg);
    stop();
  }
  return true;
}

void Task::stop(void)
{
  if (_curr->start)
  {
    if (_curr->alloc)
      free(_curr->alloc);
    dispatch();
  }
}

void Task::enqueue0(tcb_t *tcb)
{
  size_t save = System::disableAndSaveInterrupts();
  if (_tail[0])
  {
    tcb->next = _tail[0]->next;
    _tail[0]  = (_tail[0]->next = tcb);
  }
  else
  {
    tcb->next = _head;
    _tail[0]  = (_head = tcb);
  }
  System::restoreInterrupts(save);
}

void Task::enqueue1(tcb_t *tcb)
{
  size_t save = System::disableAndSaveInterrupts();
  (_tail[1] = *(_tail[1] ? &_tail[1]->next : (_tail[0] ? &_tail[0]->next : &_head)) = tcb)->next = 0;
  System::restoreInterrupts(save);
}

Task::tcb_t *Task::dequeue(void)
{
  tcb_t *tcb;
  size_t save = System::disableAndSaveInterrupts();
  if ((tcb = _head))
  {
    if (_head == _tail[0])
      _tail[0] = 0;
    else if (_head == _tail[1])
      _tail[1] = 0;
    _head = _head->next;
  }
  System::restoreInterrupts(save);
  return tcb;
}

size_t Task::current(void)
{
  return (size_t)_curr;
}

bool Task::yield(void)
{
  enqueue1(_curr);
  switching();
  return !_curr->start;
}

void Task::switching(void)
{
  if (!setjmp(_curr->context))
    dispatch();
  _curr = _disp;
}

void Task::dispatch(void)
{
  while (1)
  {
    if ((_disp = dequeue()))
      longjmp(_disp->context, 1);
#if defined(JENNIC_CHIP_FAMILY_JN516x)
    System::enableInterrupts();
    ::polling();
    System::cpuDoze();
#endif
  }
}

void Task::sleep(int32_t ms, uint8_t busyLoop)
{
  sleepMicroseconds(ms * 1000L, busyLoop);
}

void Task::sleepMicroseconds(int32_t us, uint8_t busyLoop)
{
  if (!_curr->start)
    System::delayMicroseconds(us, busyLoop);
  else
  {
    uint32 t = System::micros();
    if (us > busyLoop)
    {
      Sync sync;
      sync.init();
      sync.sleep(us -= busyLoop);
      t += us;
    }
    if (busyLoop)
    {
      while (System::micros() - t < busyLoop)
        continue;
    }
  }
}

void Task::sleepMicroseconds64(int64_t us, uint8_t busyLoop)
{
  if (!_curr->start)
    System::delayMicroseconds64(us, busyLoop);
  else
  {
    uint64 t = System::micros64();
    if (us > busyLoop)
    {
      Sync sync;
      sync.init();
      sync.sleep(us -= busyLoop);
      t += us;
    }
    if (busyLoop)
    {
      while (System::micros64() - t < busyLoop)
        continue;
    }
  }
}

void Task::delayMicroseconds(int32_t us, uint8 busyLoop)
{
  System::delayMicroseconds(us, busyLoop);
}

void Task::delayMicroseconds64(int64_t us, uint8_t busyLoop)
{
  System::delayMicroseconds64(us, busyLoop);
}

void Sync::init(void)
{
  _head = _tail = 0;
}

bool Sync::timeup(Alarm& alarm)
{
  if (((Sync *)alarm.param(0))->wakeup((size_t)alarm.param(1)))
    *(bool *)alarm.param(2) = false;
  return false;
}

bool Sync::sleep(ctrl_t& ctrl)
{
  tcb_t *tcb = (tcb_t *)current();
  (_tail = *(_tail ? &_tail->next : &_head) = tcb)->next = 0;
  System::restoreInterrupts(ctrl.save);
  ctrl.alarm.param(0, this);
  ctrl.alarm.param(1, tcb);
  ctrl.alarm.param(2, &ctrl.retval);
  ctrl.alarm.handler(timeup);
  AlarmTimer.start(ctrl.alarm);
  ++_wait;
  switching();
  --_wait;
  return ctrl.retval;
}

bool Sync::sleep(int64_t timeout)
{
  if (timeout == 0)
    return false;
  ctrl_t ctrl;
  ctrl.alarm.interval(timeout);
  ctrl.retval = true;
  ctrl.save   = System::disableAndSaveInterrupts();
  return sleep(ctrl);
}

size_t Sync::wakeup(size_t task)
{
  tcb_t *tcb = 0;
  size_t save = System::disableAndSaveInterrupts();
  if (task)
  {
    for (tcb_t **p = &_head; *p; p = &(*p)->next)
    {
      if (*p == (tcb_t *)task)
      {
        if (!(*p = (*p)->next))
          _tail = (p == &_head ? 0 : (tcb_t *)p);
        tcb = (tcb_t *)task;
        break;
      }
    }
  }
  else
    _head = ((tcb = _head) != _tail ? _head->next : _tail = 0);
  System::restoreInterrupts(save);
  if (tcb)
  {
    AlarmTimer.cancel(1, tcb);
    enqueue0(tcb);
  }
  return (size_t)tcb;
}

void Mutex::init(void)
{
  Sync::init();
  _owner = 0;
}

bool Mutex::lock(int64_t timeout)
{
  ctrl_t ctrl;
  ctrl.alarm.interval(timeout);
  ctrl.retval = true;
  ctrl.save   = System::disableAndSaveInterrupts();
  size_t task = current();
  if (!_owner)
  {
    _owner = task;
    _count = 1;
  }
  else if (_owner == task)
    ++_count;
  else if (timeout == 0)
    ctrl.retval = false;
  else
    return sleep(ctrl);
  System::restoreInterrupts(ctrl.save);
  return ctrl.retval;
}

void Mutex::unlock(void)
{
  size_t save = System::disableAndSaveInterrupts();
  if (_owner == current())
  {
    if (--_count == 0)
    {
      _owner = wakeup();
      _count = 1;
    }
  }
  System::restoreInterrupts(save);
}

void Sem::init(size_t limit, size_t count)
{
  Sync::init();
  _limit = limit;
  _count = count;
}

bool Sem::wait(size_t count, int64_t timeout)
{
  ctrl_t ctrl;
  ctrl.alarm.interval(timeout);
  ctrl.retval = true;
  ctrl.save   = System::disableAndSaveInterrupts();
  if ((count == 0) || (count > _limit))
    ctrl.retval = false;
  else if (!_wait && (_count >= count))
    _count -= count;
  else if (timeout == 0)
    ctrl.retval = false;
  else
  {
    ((tcb_t *)current())->arg = (void *)count;
    sleep(ctrl);
    while (_head)
    {
      count = (size_t)_head->arg;
      if (_count < count)
        break;
      _count -= count;
      wakeup();
    }
    return ctrl.retval;
  }
  System::restoreInterrupts(ctrl.save);
  return ctrl.retval;
}

void Sem::post(size_t count)
{
  size_t save = System::disableAndSaveInterrupts();
  if (count && (count + _count <= _limit))
  {
    _count += count;
    if (_head)
    {
      count = (size_t)_head->arg;
      if (_count >= count)
      {
        _count -= count;
        wakeup();
      }
    }
  }
  System::restoreInterrupts(save);
}

void Event::init(void)
{
  Sem::init(1);
}

bool Event::wait(int64_t timeout)
{
  return Sem::wait(1, timeout);
}

void Event::signal(void)
{
  Sem::post();
}
