/*
  timer.cpp - Timer Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "timer.h"

const Timer::TIMER Timer::TIMERS[] =
{
  Timer::TIMER_1,    /* Timer 1 */
  Timer::TIMER_2,    /* Timer 2 */
  (Timer::TIMER)-1, /* System controller */
  (Timer::TIMER)-1, /* Baseband controller */
  (Timer::TIMER)-1, /* Encryption engine */
  (Timer::TIMER)-1, /* Phy controller */
  (Timer::TIMER)-1, /* UART 0 */
  (Timer::TIMER)-1, /* UART 1 */
  Timer::TIMER_0,    /* Timer 0 */
  (Timer::TIMER)-1, /* (undefined) */
  (Timer::TIMER)-1, /* Serial Interface (2 wire) */
  (Timer::TIMER)-1, /* SPI master */
  (Timer::TIMER)-1, /* SPI slave */
  (Timer::TIMER)-1, /* Analogue peripherals */
  Timer::TIMER_3, /* Timer 3 */
  (Timer::TIMER)-1, /* Tick timer */
  Timer::TIMER_4,    /* Timer 4 */
  (Timer::TIMER)-1, /* Flash and EEPROM Controller */
  (Timer::TIMER)-1, /* (undefined) */
  (Timer::TIMER)-1, /* Infrared */
};

uint8 Timer::_diomask;

void Timer::begin(TIMER timer, CLOCK clock, bool rise, bool period, bool output, PR_HWINT_APPCALLBACK prTimerCallback)
{
  if (prTimerCallback && (rise || period))
  {
    switch (timer)
    {
      case TIMER_0: vAHI_Timer0RegisterCallback(prTimerCallback); break;
      case TIMER_1: vAHI_Timer1RegisterCallback(prTimerCallback); break;
      case TIMER_2: vAHI_Timer2RegisterCallback(prTimerCallback); break;
      case TIMER_3: vAHI_Timer3RegisterCallback(prTimerCallback); break;
      case TIMER_4: vAHI_Timer4RegisterCallback(prTimerCallback); break;
    }
    vAHI_TimerEnable(timer, clock, rise, period, output);
  }
  else
  {
    vAHI_TimerEnable(timer, clock, false, false, output);
  }
}

void Timer::end(TIMER timer)
{
  vAHI_TimerDisable(timer);
}

void Timer::clockSelect(TIMER timer, bool external, bool invert)
{
  if (timer == TIMER_0)
    vAHI_TimerClockSelect(timer, external, invert);
}

void Timer::outputs(TIMER timer, bool invert, bool input)
{
  vAHI_TimerConfigureOutputs(timer, invert, input);
}

void Timer::inputs(TIMER timer, bool invert, bool falling)
{
  if (timer == TIMER_0)
    vAHI_TimerConfigureInputs(timer, invert, falling);
}

void Timer::location(TIMER timer, bool location, bool override)
{
  vAHI_TimerSetLocation(timer, location, override);
}

void Timer::startSingleShot(TIMER timer, uint16 hi, uint16 lo)
{
  vAHI_TimerStartSingleShot(timer, hi, lo);
}

void Timer::startRepeat(TIMER timer, uint16 hi, uint16 lo)
{
  vAHI_TimerStartRepeat(timer, hi, lo);
}

void Timer::startCapture(TIMER timer)
{
  if (timer == TIMER_0)
    vAHI_TimerStartCapture(timer);
}

void Timer::startDeltaSigma(TIMER timer, uint16 hi, bool rtz)
{
  vAHI_TimerStartDeltaSigma(timer, hi, 0, rtz);
}

uint16 Timer::readCount(TIMER timer)
{
  return u16AHI_TimerReadCount(timer);
}

void Timer::readCapture(TIMER timer, uint16 *hi, uint16 *lo)
{
  if (timer == TIMER_0)
    vAHI_TimerReadCapture(timer, hi, lo);
}

void Timer::peekCapture(TIMER timer, uint16 *hi, uint16 *lo)
{
  if (timer == TIMER_0)
    vAHI_TimerReadCaptureFreeRunning(timer, hi, lo);
}

void Timer::stop(TIMER timer)
{
  vAHI_TimerStop(timer);
}

void Timer::DIOControl(TIMER timer, bool enable)
{
  uint8 mask = 0;
  switch (timer)
  {
    case TIMER_0:
      mask = DIOMASK_TIMER0_OUTPUT | DIOMASK_TIMER0_INPUT | DIOMASK_TIMER0_CAPTURE;
      break;
    case TIMER_1:
      mask = DIOMASK_TIMER1_OUTPUT;
      break;
    case TIMER_2:
      mask = DIOMASK_TIMER2_OUTPUT;
      break;
    case TIMER_3:
      mask = DIOMASK_TIMER3_OUTPUT;
      break;
    case TIMER_4:
      mask = DIOMASK_TIMER4_OUTPUT;
      break;
  }
  DIOControl(mask, enable);
}

void Timer::DIOControl(/*DIOMASK*/uint8 mask, bool enable)
{
  if (enable)
    _diomask &= ~mask;
  else
    _diomask |= mask;
  vAHI_TimerFineGrainDIOControl(_diomask);
}

Timer::FIREMASK Timer::fired(TIMER timer)
{
  // bit combination
  return (FIREMASK)u8AHI_TimerFired(timer);
}

Timer::TIMER Timer::id(uint32 device)
{
  return TIMERS[device];
}

