/*
  timer.h - Timer Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <jendefs.h>
#include <AppHardwareApi.h>

/*
   I/O Ports

            Input     Capture   Output
   -------------------------------------------
   TIMER_0: DIO8/DIO2 DIO9/DIO3 DIO10/DIO4
   TIMER_1:     -         -     DIO11/DIO5
   TIMER_2:     -         -     DIO12/DIO6/DO0
   TIMER_3:     -         -     DIO13/DIO7/DO1
   TIMER_4:     -         -     DIO17/DIO8
*/

class Timer
{
  public:

    typedef enum
    {
      TIMER_0 = E_AHI_TIMER_0, // Timer 0
      TIMER_1 = E_AHI_TIMER_1, // Timer 1
      TIMER_2 = E_AHI_TIMER_2, // Timer 2 (Infra-Red)
      TIMER_3 = E_AHI_TIMER_3, // Timer 3
      TIMER_4 = E_AHI_TIMER_4, // Timer 4
    } TIMER;

    typedef enum
    {
      CLOCK_16MHZ   = 0,
      CLOCK_8MHZ    = 1,
      CLOCK_4MHZ    = 2,
      CLOCK_2MHZ    = 3,
      CLOCK_1MHZ    = 4,
      CLOCK_500KHZ  = 5,
      CLOCK_250KHZ  = 6,
      CLOCK_125KHZ  = 7,
      CLOCK_62500HZ = 8,
      CLOCK_31250HZ = 9,
      CLOCK_15625HZ = 10,
    } CLOCK;

    static void begin(TIMER timer, CLOCK clock = CLOCK_16MHZ, bool rise = false, bool period = false, bool output = false, PR_HWINT_APPCALLBACK prTimerCallback = 0);
    static void end(TIMER timer);
    static void clockSelect(TIMER timer, bool external, bool invert);
    static void outputs(TIMER timer, bool invert, bool input = true);
    static void inputs(TIMER timer, bool invert, bool falling);
    static void location(TIMER timer, bool location, bool override);
    static void startSingleShot(TIMER timer, uint16 hi, uint16 lo);
    static void startRepeat(TIMER timer, uint16 hi, uint16 lo);
    static void startCapture(TIMER timer);
    static void startDeltaSigma(TIMER timer, uint16 hi, bool rtz);
    static uint16 readCount(TIMER timer);
    static void readCapture(TIMER timer, uint16 *hi, uint16 *lo);
    static void peekCapture(TIMER timer, uint16 *hi, uint16 *lo);
    static void stop(TIMER timer);

    typedef enum
    {
      DIOMASK_TIMER0_INPUT   = 0x01, // Timer 0 external gate/event input
      DIOMASK_TIMER0_CAPTURE = 0x02, // Timer 0 capture input
      DIOMASK_TIMER0_OUTPUT  = 0x04, // Timer 0 PWM output
      DIOMASK_TIMER1_OUTPUT  = 0x08, // Timer 1 PWM output
      DIOMASK_TIMER2_OUTPUT  = 0x10, // Timer 2 PWM output
      DIOMASK_TIMER3_OUTPUT  = 0x20, // Timer 3 PWM output
      DIOMASK_TIMER4_OUTPUT  = 0x40, // Timer 4 PWM output
    } DIOMASK;

    static void DIOControl(TIMER timer, bool enable);
    static void DIOControl(/*DIOMASK*/uint8 mask, bool enable);

    typedef enum
    {
      FIREMASK_RISE   = E_AHI_TIMER_RISE_MASK,
      FIREMASK_PERIOD = E_AHI_TIMER_PERIOD_MASK,
    } FIREMASK; // bit combination

    static FIREMASK fired(TIMER timer);
    static TIMER id(uint32 device);

  private:

    static const TIMER TIMERS[];
    static uint8 _diomask;
};
