/*
  uart.cpp - UART Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "system.h"
#include "uart.h"

Uart Serial0;
Uart Serial1;

bool Uart::begin(SPEED speed, MODE mode, FLOW flow, bool location,
  bool txOnly, /*INTMASK*/uint8 imask, PR_HWINT_APPCALLBACK prUartcallback, INTLEVEL level)
{
  // class init (constructor not work)
  if (this == &Serial0)
    _uart = UART_0;
  else if (this == &Serial1)
    _uart = UART_1;
  // Calculate Devisor and Cpb.
  uint32 u32Div = 0;
  uint32 u32Cpb;
  do
    u32Cpb = (F_PCK / ++u32Div) / speed;
  while (u32Cpb > 16);
  if ((u32Div > 255) || (u32Cpb < 4))
    return false;
  //
  if (_uart == UART_0)
    vAHI_UartSetRTSCTS(_uart, flow != FLOW_NONE);
  if (_uart == UART_1)
    vAHI_UartTxOnly(_uart, txOnly);
  vAHI_UartSetLocation(_uart, location);
  bool rv = bAHI_UartEnable(_uart, _au8TxBuf, UART_BUFFER_SIZE, _au8RxBuf, UART_BUFFER_SIZE);
  if (rv)
  {
    vAHI_UartSetBaudDivisor(_uart, u32Div);
    vAHI_UartSetClocksPerBit(_uart, u32Cpb - 1);
    vAHI_UartSetControl
    (
      _uart,
      (mode >> 3) & 1 ? E_AHI_UART_EVEN_PARITY   : E_AHI_UART_ODD_PARITY,
      (mode >> 3)     ? E_AHI_UART_PARITY_ENABLE : E_AHI_UART_PARITY_DISABLE,
      (mode &  3),
      (mode >> 2) & 1 ? E_AHI_UART_2_STOP_BITS   : E_AHI_UART_1_STOP_BIT,
      flow == FLOW_RTS_CTS_HIGH ? E_AHI_UART_RTS_HIGH : E_AHI_UART_RTS_LOW
    );
    if ((_uart == UART_0) && (flow != FLOW_NONE))
      vAHI_UartSetAutoFlowCtrl(_uart, RXLEVEL_8, flow == FLOW_RTS_CTS_HIGH, true, true);
    if (prUartcallback)
    {
      switch (_uart)
      {
        case UART_0: vAHI_Uart0RegisterCallback(prUartcallback); break;
        case UART_1: vAHI_Uart1RegisterCallback(prUartcallback); break;
      }
    }
    else
      imask = 0;
    vAHI_UartSetInterrupt
    (
      _uart,
      imask & INTMASK_MODEM,
      imask & INTMASK_STATUS,
      imask & INTMASK_TXEMPTY,
      imask & INTMASK_RXDATA,
      level
    );
  }
  return rv;
}

void Uart::end(void)
{
  vAHI_UartDisable(_uart);
}

void Uart::setBreak(bool bBreak)
{
  vAHI_UartSetBreak(_uart, bBreak);
}

void Uart::reset(bool bTxReset, bool bRxReset)
{
  vAHI_UartReset(_uart, bTxReset, bRxReset);
}

Uart::LINESTATUS Uart::lineStatus(void)
{
  // bit combination
  return (LINESTATUS)u8AHI_UartReadLineStatus(_uart);
}

Uart::MODEMSTATUS Uart::modemStatus(void)
{
  // bit combination
  return (MODEMSTATUS)(_uart == UART_0 ? u8AHI_UartReadModemStatus(_uart) : 0);
}

Uart::INTSTATUS Uart::interruptStatus(void)
{
  // bit combination
  return (INTSTATUS)u8AHI_UartReadInterruptStatus(_uart);
}

uint16 Uart::availableForWrite(void)
{
  return UART_BUFFER_SIZE - u16AHI_UartReadTxFifoLevel(_uart);
}

uint16 Uart::write(uint8 u8Data, uint32 timeout)
{
  uint32 now = System::micros();
  do
  {
    if (availableForWrite())
    {
      vAHI_UartWriteData(_uart, u8Data);
      return 1;
    }
    yield();
  }
  while (System::micros() - now < timeout);
  return 0;
}

uint16 Uart::write(const uint8 *pu8Buf, uint16 u16Len, uint32 timeout)
{
  uint16 cnt = 0;
  if (pu8Buf)
  {
    uint32 now = System::micros();
    do
    {
      if (availableForWrite())
      {
        cnt += u16AHI_UartBlockWriteData(_uart, (uint8 *)pu8Buf + cnt, u16Len - cnt);
        if (cnt >= u16Len)
          break;
        now = System::micros();
      }
      yield();
    }
    while (System::micros() - now < timeout);
  }
  return cnt;
}

uint16 Uart::available(void)
{
  return u16AHI_UartReadRxFifoLevel(_uart);
}

int Uart::read(uint32 timeout)
{
  uint32 now = System::micros();
  do
  {
    if (available())
      return u8AHI_UartReadData(_uart);
    yield();
  }
  while (System::micros() - now < timeout);
  return -1;
}

uint16 Uart::read(uint8 *pu8Buf, uint16 u16Len, uint32 timeout)
{
  uint16 cnt = 0;
  if (pu8Buf)
  {
    uint32 now = System::micros();
    do
    {
      if (available())
      {
        cnt += u16AHI_UartBlockReadData(_uart, pu8Buf + cnt, u16Len - cnt);
        if (cnt >= u16Len)
          break;
        now = System::micros();
      }
      yield();
    }
    while (System::micros() - now < timeout);
  }
  return cnt;
}
