/*
  uart.h - UART Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <jendefs.h>
#include <AppHardwareApi.h>

/*
   I/O Ports

           TXD         RXD        CTS        RTS
   ---------------------------------------------------
   UART_0: DIO6/DIO14  DIO7/DIO15 DIO4/DIO12 DIO5/DIO13
   UART_1: DIO14/DIO11 DIO15/DIO9     -          -
*/

#ifndef UART_BUFFER_SIZE
#define UART_BUFFER_SIZE 128 // Size of TX/RX Buffer, in range 16 to 2047 bytes
#endif

class Uart;

extern Uart Serial0;
extern Uart Serial1;

class Uart
{
  public:

    typedef enum
    {
      UART_0 = E_AHI_UART_0,
      UART_1 = E_AHI_UART_1,
    } UART;

    typedef enum
    {
      SPEED_4800    = 4800,
      SPEED_9600    = 9600,
      SPEED_14400   = 14400,
      SPEED_19200   = 19200,
      SPEED_38400   = 38400,
      SPEED_57600   = 57600,
      SPEED_115200  = 115200,
      SPEED_230400  = 230400,
      SPEED_460800  = 460800,
      SPEED_500000  = 500000,
      SPEED_921600  = 921600,
      SPEED_1M      = 1000000,
      SPEED_2M      = 2000000,
      SPEED_4M      = 4000000,
    } SPEED;

    typedef enum
    {
      FLOW_NONE,
      FLOW_RTS_CTS_LOW,
      FLOW_RTS_CTS_HIGH,
    } FLOW;

    typedef enum
    {
      MODE_5N1 = 0,
      MODE_6N1 = 1,
      MODE_7N1 = 2,
      MODE_8N1 = 3,
      MODE_5N2 = 4,
      MODE_6N2 = 5,
      MODE_7N2 = 6,
      MODE_8N2 = 7,
      MODE_5E1 = 8,
      MODE_6E1 = 9,
      MODE_7E1 = 10,
      MODE_8E1 = 11,
      MODE_5E2 = 12,
      MODE_6E2 = 13,
      MODE_7E2 = 14,
      MODE_8E2 = 15,
      MODE_5O1 = 16,
      MODE_6O1 = 17,
      MODE_7O1 = 18,
      MODE_8O1 = 19,
      MODE_5O2 = 20,
      MODE_6O2 = 21,
      MODE_7O2 = 22,
      MODE_8O2 = 23,
    } MODE;

    typedef enum
    {
      INTMASK_MODEM   = 0x01, // CTS change detected for UART0
      INTMASK_STATUS  = 0x02, // break indication, framing error, parity error or over-run
      INTMASK_TXEMPTY = 0x04, // Transmit FIFO empty
      INTMASK_RXDATA  = 0x08, // receive data available
    } INTMASK;

    typedef enum
    {
      INTLEVEL_1  = E_AHI_UART_FIFO_LEVEL_1,
      INTLEVEL_4  = E_AHI_UART_FIFO_LEVEL_4,
      INTLEVEL_8  = E_AHI_UART_FIFO_LEVEL_8,
      INTLEVEL_14 = E_AHI_UART_FIFO_LEVEL_14,
    } INTLEVEL;

    bool begin(SPEED speed = SPEED_115200, MODE mode = MODE_8N1, FLOW flow = FLOW_NONE, bool location = false,
        bool txOnly = false, /*INTMASK*/uint8 imask = 0, PR_HWINT_APPCALLBACK prUartcallback = 0, INTLEVEL level = INTLEVEL_1);
    void end(void);

    typedef enum
    {
      RXLEVEL_8  = E_AHI_UART_FIFO_ARTS_LEVEL_8,
      RXLEVEL_11 = E_AHI_UART_FIFO_ARTS_LEVEL_11,
      RXLEVEL_13 = E_AHI_UART_FIFO_ARTS_LEVEL_13,
      RXLEVEL_15 = E_AHI_UART_FIFO_ARTS_LEVEL_15,
    } RXLEVEL;

    void setBreak(bool bBreak);
    void reset(bool bTxReset = true, bool bRxReset = true);

    typedef enum
    {
      LINESTATUS_ERROR   = E_AHI_UART_LS_ERROR,
      LINESTATUS_TXREADY = E_AHI_UART_LS_TEMT,
      LINESTATUS_TXEMPTY = E_AHI_UART_LS_THRE,
      LINESTATUS_BI      = E_AHI_UART_LS_BI,
      LINESTATUS_FE      = E_AHI_UART_LS_FE,
      LINESTATUS_PE      = E_AHI_UART_LS_PE,
      LINESTATUS_OE      = E_AHI_UART_LS_OE,
      LINESTATUS_RXREADY = E_AHI_UART_LS_DR,
    } LINESTATUS; // bit combination

    LINESTATUS lineStatus(void);

    typedef enum
    {
      MODEMSTATUS_DCTS = E_AHI_UART_MS_DCTS, // CTS input has changed.
      MODEMSTATUS_CTS  = E_AHI_UART_MS_CTS,  // ‘1’ indicates CTS is high, ‘0’ indicates CTS is low
    } MODEMSTATUS; // bit combination
    MODEMSTATUS modemStatus(void);

    typedef enum
    {
      INTSTATUS_RXLINE  = E_AHI_UART_INT_RXLINE,
      INTSTATUS_RXDADA  = E_AHI_UART_INT_RXDATA,
      INTSTATUS_TIMEOUT = E_AHI_UART_INT_TIMEOUT,
      INTSTATUS_TX      = E_AHI_UART_INT_TX,
      INTSTATUS_MODEM   = E_AHI_UART_INT_MODEM,
    } INTSTATUS; // bit combination

    INTSTATUS interruptStatus(void);
    uint16 availableForWrite(void);
    uint16 write(uint8 u8Data, uint32 timeout = 1000);
    uint16 write(const uint8 *pu8Buf, uint16 u16Len, uint32 timeout = 1000);
    uint16 available(void);
    int read(uint32 timeout = 1000);
    uint16 read(uint8 *pu8Buf, uint16 u16Len, uint32 timeout = 1000);

  private:

    UART  _uart;
    uint8 _au8TxBuf[UART_BUFFER_SIZE];
    uint8 _au8RxBuf[UART_BUFFER_SIZE];
};
