/*
  wire.cpp - Two Wire Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "wire.h"

SWire::SWIRECALLBACK SWire::_prSWireCallback;
uint16 SWire::_rdcount;
uint8  SWire::_rdbuffer[];
uint16 SWire::_wrindex;
uint16 SWire::_wrcount;
uint8  SWire::_wrbuffer[];

void Wire::begin(SPEED speed, bool bPulseSuppressionEnable)
{
  vAHI_SiMasterConfigure(bPulseSuppressionEnable, false, F_PCK / speed / 5 - 1);
}

void Wire::end(void)
{
  vAHI_SiMasterDisable();
}

void Wire::location(bool bLocation)
{
  // DIOs on which interface will operate:
  //   TRUE  - DIO16-17
  //   FALSE - DIO14-15 (default)
  vAHI_SiSetLocation(bLocation);
}

bool Wire::write(uint16 slave, uint8 u8Cmd, const uint8 *pu8Buf, uint16 u16Len, bool stop)
{
  if (!pu8Buf)
    u16Len = 0;
  if (!start(slave, false, false))
    return false;
  while (1)
  {
    vAHI_SiMasterWriteData8(u8Cmd);
    if (!command(!u16Len && stop ? CMD_WRITE_STOP : CMD_WRITE))
      return false;
    if (u16Len-- == 0)
      break;
    u8Cmd = *pu8Buf++;
  }
  return true;
}

bool Wire::write(uint16 slave, const uint8 *pu8Buf, uint16 u16Len, bool stop)
{
  if (!pu8Buf)
    u16Len = 0;
  if (!start(slave, !u16Len && stop, false))
    return false;
  while (u16Len)
  {
    vAHI_SiMasterWriteData8(*pu8Buf++);
    if (!command(!--u16Len && stop ? CMD_WRITE_STOP : CMD_WRITE))
      return false;
  }
  return true;
}

bool Wire::read(uint16 slave, uint8 *pu8Buf, uint16 u16Len, bool stop)
{
  if (!pu8Buf)
    u16Len = 0;
  if (!start(slave, !u16Len && stop, true))
    return false;
  while (u16Len)
  {
    if (!command(!--u16Len && stop ? CMD_READ_STOP : CMD_READ))
      return false;
    *pu8Buf++ = u8AHI_SiMasterReadData8();
  }
  return true;
}

bool Wire::writeAndRead(uint16 slave, uint8 u8Cmd, const uint8 *pu8Write, uint16 u16WriteLen, uint8 *pu8Read, uint16 u16ReadLen, bool stop)
{
  return write(slave, u8Cmd, pu8Write, u16WriteLen, false) && read(slave, pu8Read, u16ReadLen, stop);
}

bool Wire::writeAndRead(uint16 slave, const uint8 *pu8Write, uint16 u16WriteLen, uint8 *pu8Read, uint16 u16ReadLen, bool stop)
{
  return write(slave, pu8Write, u16WriteLen, false) && read(slave, pu8Read, u16ReadLen, stop);
}

void Wire::bus(CMD cmd)
{
  bAHI_SiMasterSetCmdReg
  (
    cmd & (CMD_START - CMD_WRITE) ? E_AHI_SI_START_BIT   : E_AHI_SI_NO_START_BIT,   // bSetSTA,
    cmd &  CMD_STOP               ? E_AHI_SI_STOP_BIT    : E_AHI_SI_NO_STOP_BIT,    // bSetSTO,
    cmd &  CMD_READ               ? E_AHI_SI_SLAVE_READ  : E_AHI_SI_NO_SLAVE_READ,  // bSetRD,
    cmd &  CMD_WRITE              ? E_AHI_SI_SLAVE_WRITE : E_AHI_SI_NO_SLAVE_WRITE, // bSetWR,
    cmd == CMD_READ               ? E_AHI_SI_SEND_ACK    : E_AHI_SI_SEND_NACK,      // bSetAckCtrl
    E_AHI_SI_NO_IRQ_ACK                                                             // bSetIACK
  );
}

bool Wire::command(CMD cmd)
{
  bus(cmd);
  while (bAHI_SiMasterPollTransferInProgress())
    continue;
  if (!bAHI_SiMasterPollArbitrationLost() && !bAHI_SiMasterCheckRxNack())
    return true;
  if (!(cmd & CMD_STOP))
    bus(CMD_STOP);
  return false;
}

bool Wire::start(uint16 slave, bool stop, bool read)
{
  if (slave & ~0x7F)
  {
    vAHI_SiMasterWriteSlaveAddr(0x78 | ((slave >> 8) & 0x03), read);
    if (command(CMD_START))
    {
      vAHI_SiMasterWriteData8(slave);
      if (command(stop ? CMD_WRITE_STOP : CMD_WRITE))
        return true;
    }
  }
  else
  {
    vAHI_SiMasterWriteSlaveAddr(slave & 0x7F, read);
    if (command(stop ? CMD_START_STOP : CMD_START))
      return true;
  }
  return false;
}

void SWire::begin(SWIRECALLBACK prSWireCallback, uint16 u16SlaveAddress, bool bPulseSuppressionEnable)
{
  _prSWireCallback = prSWireCallback;
  _rdcount = 0;
  _wrcount = 0;
  _wrindex = 0;
  vAHI_SiRegisterCallback(SiCallback);
  vAHI_SiSlaveConfigure(u16SlaveAddress & 0x3FF, u16SlaveAddress & ~0x7F,
    bPulseSuppressionEnable, IMASK_RR | IMASK_RTKN | IMASK_WA | IMASK_LAST | IMASK_ERROR, false);
}

void SWire::end(void)
{
  vAHI_SiSlaveDisable();
}

void SWire::write(uint8 u8Out)
{
  if (_wrcount < WIRES_BUFFER_SIZE)
    _wrbuffer[_wrcount++] = u8Out;
}

void SWire::write(uint8 *pu8Out, uint16 u16Len)
{
  while (u16Len--)
    write(*pu8Out++);
}

void SWire::SiCallback(uint32 u32Device, uint32 u32ItemBitmap)
{
  // reset
  if (u32ItemBitmap & (IMASK_RTKN | IMASK_ERROR))
  {
    _rdcount = 0;
    _wrcount = 0;
    _wrindex = 0;
  }
  // write
  if (u32ItemBitmap & IMASK_RR)
  {
    if (_wrindex < _wrcount)
      vAHI_SiSlaveWriteData8(_wrbuffer[_wrindex++]);
  }
  // read
  if (u32ItemBitmap & IMASK_WA)
  {
    uint8 b = u8AHI_SiSlaveReadData8();
    if (_rdcount < WIRES_BUFFER_SIZE)
      _rdbuffer[_rdcount++] = b;
  }
  // callback
  if ((u32ItemBitmap & IMASK_LAST) && _prSWireCallback)
    _prSWireCallback(_rdbuffer, _rdcount);
}
