/*
  wire.h - Two Wire Library for NXP-JN516x

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <jendefs.h>
#include <AppHardwareApi.h>

/*
   I/O Ports

     SCK: DIO14/DIO16
     SDA: DIO15/DIO17
*/

#define WIRE_TENBIT_ADDRESS 0x8000

#ifndef WIRES_BUFFER_SIZE
#define WIRES_BUFFER_SIZE 128
#endif

class Wire
{
  public:

    typedef enum
    {
      SPEED_100KHZ = 100000, // Standard
      SPEED_400KHZ = 400000, // Fast
      SPEED_800KHZ = 800000, // Fast-Plus (Max 1.0MHz)
    } SPEED;

    static void begin(SPEED speed = SPEED_100KHZ, bool bPulseSuppressionEnable = true);
    static void end(void);
    static void location(bool bLocation);
    static bool write(uint16 slave, const uint8 *pu8Buf, uint16 u16Len, bool stop = true);
    static bool read(uint16 slave, uint8 *pu8Buf, uint16 u16Len, bool stop = true);
    static bool writeAndRead(uint16 slave, const uint8 *pu8Write, uint16 u16WriteLen, uint8 *pu8Read, uint16 u16ReadLen, bool stop = true);
    static bool write(uint16 slave, uint8 u8Cmd, const uint8 *pu8Buf, uint16 u16Len, bool stop = true);
    static bool writeAndRead(uint16 slave, uint8 u8Cmd, const uint8 *pu8Write, uint16 u16WriteLen, uint8 *pu8Read, uint16 u16ReadLen, bool stop = true);

  private:

    typedef enum
    {
      // Start Stop Read Write Resulting Instruction to SI Bus
      CMD_IDLE       = 0b0000, // No active command (idle)
      CMD_START      = 0b1001, // Start followed by Write
      CMD_START_STOP = 0b1101, // Start followed by Write followed by Stop
      CMD_READ_STOP  = 0b0110, // Read followed by Stop
      CMD_WRITE_STOP = 0b0101, // Write followed by Stop
      CMD_WRITE      = 0b0001, // Write only
      CMD_READ       = 0b0010, // Read only
      CMD_STOP       = 0b0100, // Stop only
    } CMD;

    static void bus(CMD cmd);
    static bool command(CMD cmd);
    static bool start(uint16 slave, bool stop, bool read);
};

class SWire
{
  public:

    typedef void (*SWIRECALLBACK)(const uint8 *pu8Buf, uint16 u16Len);

    typedef enum
    {
      IMASK_RR    = E_AHI_SIS_DATA_RR_MASK,   // Data buffer must be written with data to be read by SI master
      IMASK_RTKN  = E_AHI_SIS_DATA_RTKN_MASK, // Data taken from buffer by SI master - buffer free for next data
      IMASK_WA    = E_AHI_SIS_DATA_WA_MASK,   // Data buffer contains data from SI master to be read by SI slave
      IMASK_LAST  = E_AHI_SIS_LAST_DATA_MASK, // Last data transferred (end of burst)
      IMASK_ERROR = E_AHI_SIS_ERROR_MASK,     // I2C protocol error
    } IMASK;

    static void begin(SWIRECALLBACK prSWireCallback, uint16 u16SlaveAddress, bool bPulseSuppressionEnable = true);
    static void end(void);
    static void write(uint8 u8Out);
    static void write(uint8 *pu8Out, uint16 u16Len);

  private:

    static void SiCallback(uint32 u32Device, uint32 u32ItemBitmap);

    static SWIRECALLBACK _prSWireCallback;
    static uint16 _rdcount;
    static uint8  _rdbuffer[WIRES_BUFFER_SIZE];
    static uint16 _wrindex;
    static uint16 _wrcount;
    static uint8  _wrbuffer[WIRES_BUFFER_SIZE];
};
