/*
  wpan.cpp - Wireless Personal Area Network Library for NXP-JN516x (IEEE802.15.4)

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <string.h>
#include <PeripheralRegs.h>
#include "jn516x.h"
#include "crc32.h"
#include "wpan.h"

WPan::tsNetwork WPan::sNetwork;

#if WPAN_SECURITY_ENABLE
static const uint8 SECURITY_KEY_DATA[] = WPAN_SECURITY_KEY_DATA;
#endif

bool WPan::begin(uint16 u16Addr, uint16 u16PanId, CHANNEL u8Channel)
{
  if (!setShortAddress(u16Addr))
    return false;
  sNetwork.u16PanId = u16PanId;
  void *s_pvMac = pvAppApiGetMacHandle();
  /* Set Pan ID in PIB (also sets match register in hardware) */
  MAC_vPibSetPanId(s_pvMac, u16PanId);
  /* Enable receiver to be on when idle */
  MAC_vPibSetRxOnWhenIdle(s_pvMac, true, false);
  /* set current channel */
  eAppApiPlmeSet((PHY_PibAttr_e)PHY_PIB_ATTR_CURRENT_CHANNEL, u8Channel);

  sNetwork.psTransmit = 0;
  for (uint32 i = 0; i < sizeof(sNetwork.sTransmitData) / sizeof(sNetwork.sTransmitData[0]); ++i)
  {
    sNetwork.sTransmitData[i].next = sNetwork.psTransmit;
    sNetwork.psTransmit = sNetwork.sTransmitData + i;
    sNetwork.psTransmit->data.uParam.sReqData.sFrame.sDstAddr.uAddr.u16Short = WPAN_MAX_DEVICES;
  }
  return true;
}

void WPan::transmit(uint16 u16Addr, const uint8 *pu8Data, uint8 u8Len, uint8 u8Handle)
{
  MAC_McpsSyncCfm_s sMcpsSyncCfm;

  tsSeqNb *psSeqNb = getSequenceNumber(u16Addr);
  if (!psSeqNb || (u8Len > WPAN_MAX_PAYLOAD_SIZE))
    return;

  uint8 u8Tx = ++psSeqNb->u8Tx;
  u8Tx = (u16Addr == WPAN_ADDR_BROADCAST ? u8Tx | 0x80 : u8Tx & ~0x80);
  tsTransmit *psTransmit = getTransmitData(u16Addr);
  /* Create frame transmission request */
  psTransmit->data.u8Type = MAC_MCPS_REQ_DATA;
  psTransmit->data.u8ParamLength = sizeof(MAC_McpsReqData_s);
  /* Set handle so we can match confirmation to request */
  psTransmit->data.uParam.sReqData.u8Handle = u8Handle;
  /* Use short address for source */
  vSetMacAddr(psTransmit->data.uParam.sReqData.sFrame.sSrcAddr, sNetwork.u16Addr);
  /* Use short address for destination */
  vSetMacAddr(psTransmit->data.uParam.sReqData.sFrame.sDstAddr, u16Addr);
  /* Frame requires ack but not security, indirect Transmit or GTS */
  psTransmit->data.uParam.sReqData.sFrame.u8TxOptions =
    u8Handle && (u16Addr != WPAN_ADDR_BROADCAST) ? MAC_TX_OPTION_ACK : 0;
  /* Copy Data */
  uint8 *pu8Payload = psTransmit->data.uParam.sReqData.sFrame.au8Sdu;
  *pu8Payload++ = u8Tx;
  if (pu8Data && u8Len)
  {
#if WPAN_SECURITY_ENABLE
    uint8 data[128];
    memcpy(&data[WPAN_SECURITY_AES_MODE], pu8Data, u8Len);
    pu8Data = data;
    AES::begin(WPAN_SECURITY_AES_MODE, (uint8 *)SECURITY_KEY_DATA, sizeof(SECURITY_KEY_DATA));
    AES::encryption(&data[WPAN_SECURITY_AES_MODE], u8Len, u8Tx, (uint8 *)&sNetwork.u16Addr, sizeof(sNetwork.u16Addr), data);
    u8Len += WPAN_SECURITY_AES_MODE;
#endif
    for (uint8 i = 0; i < u8Len; i++)
      *pu8Payload++ = *pu8Data++;
  }
  /* Set frame length */
  psTransmit->data.uParam.sReqData.sFrame.u8SduLength = pu8Payload - psTransmit->data.uParam.sReqData.sFrame.au8Sdu;
  /* Request Transmit */
  vAppApiMcpsRequest(&psTransmit->data, &sMcpsSyncCfm);
}

void WPan::reTransmit(uint16 u16Addr)
{
  MAC_McpsSyncCfm_s sMcpsSyncCfm;

  tsTransmit *psTransmit = getTransmitData(u16Addr, false);
  if (psTransmit)
    vAppApiMcpsRequest(&psTransmit->data, &sMcpsSyncCfm);
}

void WPan::handle(void)
{
  MAC_MlmeDcfmInd_s *psMlmeInd;
  MAC_McpsDcfmInd_s *psMcpsInd;
  AppQApiHwInd_s*psAHI_Ind;

  /* Check for anything on the MLME upward queue */
  if ((psMlmeInd = psAppQApiReadMlmeInd()) != NULL)
  {
    vProcessIncomingMlme(psMlmeInd);
    vAppQApiReturnMlmeIndBuffer(psMlmeInd);
  };

  /* Check for anything on the MCPS upward queue */
  if ((psMcpsInd = psAppQApiReadMcpsInd()) != NULL)
  {
    vProcessIncomingMcps(psMcpsInd);
    vAppQApiReturnMcpsIndBuffer(psMcpsInd);
  };

  /* Check for anything on the AHI upward queue */
  if ((psAHI_Ind = psAppQApiReadHwInd()) != NULL)
  {
    vProcessIncomingHwEvent(psAHI_Ind);
    vAppQApiReturnHwIndBuffer(psAHI_Ind);
  };
}

bool WPan::setShortAddress(uint16 u16Addr)
{
  if ((u16Addr == WPAN_ADDR_BROADCAST) || !getSequenceNumber(u16Addr))
    return false;
  sNetwork.u16Addr = u16Addr;
  /* Set short address in PIB (also sets match registers in hardware) */
  MAC_vPibSetShortAddr(pvAppApiGetMacHandle(), u16Addr);
  return true;
}

uint16 WPan::getShortAddress(void)
{
  return sNetwork.u16Addr;
}

void WPan::getMacAddress(MAC_ExtAddr_s &addr)
{
  tsExtAddr  mac;
  vMMAC_GetMacAddress(&mac);
  addr.u32L = mac.u32L;
  addr.u32H = mac.u32H;
}

void WPan::transmissionPower(TXPOWER txpower)
{
  // n input range of -32 to 31 dBm
  // For JN5168, JN5164 and JN5161, it is mapped to one of four levels:
  //   -32, -20, -9 and 0 dBm
  // For JN5169, it is mapped to the nearest of 26 levels in the range:
  //   -32 to 10 dBm
  eAppApiPlmeSet((PHY_PibAttr_e)PHY_PIB_ATTR_TX_POWER, txpower);
#if E_AHI_ADC_SRC_ADC_5
  vAHI_RadioSetReducedInputPower(txpower <= 0);
#endif
}

void WPan::vProcessIncomingMlme(MAC_MlmeDcfmInd_s *psMlmeInd)
{
}

void WPan::vProcessIncomingMcps(MAC_McpsDcfmInd_s *psMcpsInd)
{
  /* Only handle incoming data events one device has been started as a coordinator */
  switch(psMcpsInd->u8Type)
  {
    /* Incoming data frame */
    case MAC_MCPS_IND_DATA:
      vHandleMcpsDataInd(psMcpsInd->uParam.sIndData.sFrame);
      break;
    /* Incoming acknowledgement or ack timeout */
    case MAC_MCPS_DCFM_DATA:
      vHandleMcpsDataDcfm(psMcpsInd->uParam.sDcfmData);
      break;
  }
}

void WPan::vHandleMcpsDataInd(MAC_RxFrameData_s& sFrame)
{
  if ((sFrame.sSrcAddr.u8AddrMode == 2) && (sFrame.u8SduLength > 0))
  {
    uint16 u16Addr = sFrame.sSrcAddr.uAddr.u16Short;
    tsSeqNb *psSeqNb = getSequenceNumber(u16Addr);
    if (psSeqNb)
    {
      uint8 *pu8Sdu  = sFrame.au8Sdu + 1;
      uint8 u8SduLen = sFrame.u8SduLength - 1;
#if WPAN_SECURITY_ENABLE
      if (u8SduLen)
      {
        uint8  data[128];
        memcpy(data, pu8Sdu, u8SduLen);
        pu8Sdu = &data[WPAN_SECURITY_AES_MODE];
        u8SduLen -= WPAN_SECURITY_AES_MODE;
        AES::begin(WPAN_SECURITY_AES_MODE, (uint8 *)SECURITY_KEY_DATA, sizeof(SECURITY_KEY_DATA));
        if (!AES::decryption(pu8Sdu, u8SduLen, sFrame.au8Sdu[0], (uint8 *)&u16Addr, sizeof(u16Addr), data))
          return; // security failed.
      }
#endif
      bool broadcast = sFrame.sDstAddr.uAddr.u16Short == WPAN_ADDR_BROADCAST;
      tsTransmit *psTransmit = !broadcast ? getTransmitData(u16Addr, false) : 0;
      if (psSeqNb->u8Rx != sFrame.au8Sdu[0])
      {
        /* new packet */
        psSeqNb->u8Rx = sFrame.au8Sdu[0];
        /* remove transmit data */
        removeTransmitData(psTransmit);
        /* process received packet */
        if (!broadcast && ((u16Addr == JN516x::ADDR_OTA_BRIDGE) || (u16Addr == JN516x::ADDR_OTA_BRIDGE2)))
          OTARecieveCallback(u16Addr, pu8Sdu, u8SduLen);
        else if (WPanRecieveCallback)
          WPanRecieveCallback(u16Addr, pu8Sdu, u8SduLen, broadcast);
      }
      else if (psTransmit)
      {
        /* If it is not a broadcast packet, the last transmitted data will be retransmitted. */
        MAC_McpsSyncCfm_s sMcpsSyncCfm;
        vAppApiMcpsRequest(&psTransmit->data, &sMcpsSyncCfm);
      }
    }
  }
}

void WPan::vHandleMcpsDataDcfm(MAC_McpsCfmData_s& sDcfmData)
{
  OTATransmitCallback(sDcfmData.u8Handle, sDcfmData.u8Status);
  if (WPanTransmitCallback)
    WPanTransmitCallback(sDcfmData.u8Handle, sDcfmData.u8Status);
}

void WPan::vProcessIncomingHwEvent(AppQApiHwInd_s *psAHI_Ind)
{
  if (WPanHWEventCallback)
    WPanHWEventCallback(psAHI_Ind->u32DeviceId, psAHI_Ind->u32ItemBitmap);
}

WPan::tsTransmit *WPan::getTransmitData(uint16 u16Addr, bool alloc)
{
  tsTransmit **p, *q;
  for (p = &sNetwork.psTransmit; *p; p = &(*p)->next)
  {
    if ((alloc && !(*p)->next) || ((*p)->data.uParam.sReqData.sFrame.sDstAddr.uAddr.u16Short == u16Addr))
    {
      *p = (q = *p)->next;
      q->next = sNetwork.psTransmit;
      sNetwork.psTransmit = q;
      return q;
    }
  }
  return 0;
}

void WPan::removeTransmitData(tsTransmit *psTransmit)
{
  tsTransmit **p;
  if (psTransmit)
  {
    for (p = &sNetwork.psTransmit;; p = &(*p)->next)
    {
      if (*p == psTransmit)
      {
        *p = psTransmit->next;
        psTransmit->next = 0;
        psTransmit->data.uParam.sReqData.sFrame.sDstAddr.uAddr.u16Short = WPAN_MAX_DEVICES;
      }
      if (!(*p)->next)
      {
        (*p)->next = psTransmit;
        break;
      }
    }
  }
}

WPan::tsSeqNb *WPan::getSequenceNumber(uint16 u16Addr)
{
  return u16Addr < WPAN_MAX_DEVICES ? sNetwork.sSeqNb + u16Addr
    : (u16Addr >= (uint16)-WPAN_RSV_ADDRESS ? sNetwork.sSeqNb + WPAN_MAX_DEVICES + WPAN_RSV_ADDRESS + (int16)u16Addr : 0);
}

void WPan::vSetMacAddr(MAC_Addr_s& sAddr, uint16 u16Addr)
{
  sAddr.u8AddrMode = 2;
  sAddr.u16PanId   = sNetwork.u16PanId;
  sAddr.uAddr.u16Short = u16Addr;
}

void WPan::write32(uint8 *buf, uint32 data)
{
  buf[3] = (uint8)(data >>  0);
  buf[2] = (uint8)(data >>  8);
  buf[1] = (uint8)(data >> 16);
  buf[0] = (uint8)(data >> 24);
}

uint32 WPan::read32(const uint8 *buf)
{
  return
    ((uint32)buf[0] << 24) |
    ((uint32)buf[1] << 16) |
    ((uint16)buf[2] <<  8) |
    ((uint8 )buf[3] <<  0);
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"

void WPan::OTARecieveCallback(uint16 u16Addr, const uint8 *pu8Data, uint8 u8Len)
{
#if defined(JN5179)
  JN516x::tsFlashHeader_JN5179 sFlashHeader;
#else
  JN516x::tsFlashHeader sFlashHeader;
#endif
  uint32 u32Remap;
  uint8  u8Sector;
  uint8  u8Ack[16];
  JN516x::ota_request_t&  request  = *(JN516x::ota_request_t  *)pu8Data;
  JN516x::ota_response_t& response = *(JN516x::ota_response_t *)u8Ack;
  response.command = request.command;
  response.status  = JN516x::OTA_STATUS_OK;
  uint8  u8AckLen  = offsetof(__typeof__(response), args);
  uint8  u8Handle  = 0;
  switch (request.command)
  {
    case JN516x::OTA_UPDATE_CHIPVER:
      u32Remap = u32REG_SysRead(REG_SYS_FLASH_REMAP);
      u8Sector = JN516x::chipFlashSectors(CHIP_VERSION) - 1;
      if (u8Sector < 8)
        vREG_SysWrite(REG_SYS_FLASH_REMAP , (u32Remap & ~(0x0F << ((u8Sector & 7) << 2))) | (u8Sector << ((u8Sector & 7) << 2)));
#if defined(JN5169) || defined(JN5179)
      else
      {
        u32Remap = u32REG_SysRead(REG_SYS_FLASH_REMAP2);
        vREG_SysWrite(REG_SYS_FLASH_REMAP2, (u32Remap & ~(0x0F << ((u8Sector & 7) << 2))) | (u8Sector << ((u8Sector & 7) << 2)));
      }
#endif
      if (!FlashInit())
        response.status = JN516x::OTA_STATUS_SDK_ERROR;
      else if (!bAHI_FullFlashRead(FLASH_SECTOR_OFFSET(u8Sector), sizeof(sFlashHeader), (uint8 *)&sFlashHeader))
        response.status = JN516x::OTA_STATUS_SDK_ERROR;
      else
      {
        // check OTA-BOOT
        if (u8Sector == (u32Remap & 0x0F))
          sNetwork.bOtaValid = true;
        else if (sFlashHeader.imageSize <= FLASH_SECTOR_OFFSET(u8Sector - (u32Remap & 0x0F)))
        {
          sNetwork.bOtaValid = false;
          sNetwork.bOtaValid  = memcmp(sFlashHeader.sBIR.magicNumber, JN516x::MAGICNUMBER, sizeof(JN516x::MAGICNUMBER)) == 0;
#if defined(JN5179)
          sNetwork.bOtaValid |= memcmp(sFlashHeader.sBIR.magicNumber, JN516x::MAGICNUMBER_JN5179, sizeof(JN516x::MAGICNUMBER)) == 0;
#endif
          sNetwork.bOtaValid &= sFlashHeader.sBIR.status == 0x01;
          sNetwork.bOtaValid &= sFlashHeader.sBIR.applicationID == 0x0000;
        }
        write32((uint8_t *)&response.args.device.version, CHIP_VERSION);
        response.args.device.otaboot = sNetwork.bOtaValid;
        u8AckLen += sizeof(response.args.device);
      }
      break;

    case JN516x::OTA_UPDATE_BLDMODE:
      u8Sector = ((u8Len > offsetof(__typeof__(request), args)) && (request.args.mode == 0xBF) ? 0xBF : 0xFF);
      if (sNetwork.bOtaValid || (u8Sector == 0xBF))
        sNetwork.u8OtaHandle = u8Handle = u8Sector;
      else
        response.status = JN516x::OTA_STATUS_BLD_ERROR;
      break;

    case JN516x::OTA_UPDATE_START:
      u8Sector = u32REG_SysRead(REG_SYS_FLASH_REMAP) & 0x0F;
      if (u8Sector && (u8Sector != JN516x::chipFlashSectors(CHIP_VERSION) - 1))
        response.status = JN516x::OTA_STATUS_BLD_ERROR;
      else if (!FlashInit())
        response.status = JN516x::OTA_STATUS_SDK_ERROR;
      else
      {
        if (u8Sector)
        {
          vREG_SysWrite(REG_SYS_FLASH_REMAP , 0x65432100U | (u32REG_SysRead(REG_SYS_FLASH_REMAP) & 0x0F));
#if defined(JN5169) || defined(JN5179)
          vREG_SysWrite(REG_SYS_FLASH_REMAP2, 0xEDCBA987U);
#endif
          sNetwork.u32OtaStart = FLASH_SECTOR_OFFSET(1);
        }
        else
        {
          vREG_SysWrite(REG_SYS_FLASH_REMAP , 0x76543210U);
#if defined(JN5169) || defined(JN5179)
          vREG_SysWrite(REG_SYS_FLASH_REMAP2, 0xFEDCBA98U);
#endif
          sNetwork.u32OtaStart = FLASH_SECTOR_OFFSET(JN516x::chipFlashSectors(CHIP_VERSION) - 1);
        }
        sNetwork.u32OtaAddr  = sNetwork.u32OtaStart;
        sNetwork.u8OtaErase  = 0xFF;
        sNetwork.u8OtaHandle = 0;
        CRC32::init();
      }
      break;

    case JN516x::OTA_UPDATE_PROGRAM:
      pu8Data = request.args.data;
      u8Len  -= offsetof(__typeof__(request), args.data);
      CRC32::update(pu8Data, u8Len);
      if (sNetwork.u32OtaAddr == sNetwork.u32OtaStart)
      {
        memcpy(sNetwork.u8OtaHeader, pu8Data, sizeof(sNetwork.u8OtaHeader));
        sNetwork.u32OtaAddr += sizeof(sNetwork.u8OtaHeader);
        pu8Data += sizeof(sNetwork.u8OtaHeader);
        u8Len   -= sizeof(sNetwork.u8OtaHeader);
      }
      u8Sector = (sNetwork.u32OtaAddr + (u8Len ? u8Len - 1 : 0)) >> 15;
      if (sNetwork.u8OtaErase != u8Sector)
      {
        sNetwork.u8OtaErase = u8Sector;
        if (!bAHI_FlashEraseSector(u8Sector))
        {
          response.status = JN516x::OTA_STATUS_SDK_ERROR;
          break;
        }
      }
      if (!bAHI_FullFlashProgram(sNetwork.u32OtaAddr, u8Len, (uint8 *)pu8Data))
        response.status = JN516x::OTA_STATUS_SDK_ERROR;
      else
        sNetwork.u32OtaAddr += u8Len;
      break;

    case JN516x::OTA_UPDATE_END:
      if (CRC32::result() == read32((uint8_t *)&request.args.crc32))
        sNetwork.u8OtaHandle = u8Handle = 0xFE;
      else
      {
        response.status = JN516x::OTA_STATUS_CRC_ERROR;
        write32((uint8_t *)&response.args.crc32, CRC32::result());
        u8AckLen += sizeof(response.args.crc32);
      }
      break;

    default:
      if (WPanRecieveCallback)
      {
        WPanRecieveCallback(u16Addr, pu8Data, u8Len, false);
        return;
      }
      response.status = JN516x::OTA_STATUS_ARG_ERROR;
      break;
  }
  transmit(u16Addr, (uint8_t *)&response, u8AckLen, u8Handle);
}

void WPan::OTATransmitCallback(uint8 u8Handle, uint8 u8Status)
{
  if (sNetwork.u8OtaHandle && (sNetwork.u8OtaHandle == u8Handle))
  {
    sNetwork.u8OtaHandle = 0;
    if (u8Status == 0)
    {
      switch (u8Handle)
      {
        case 0xFF:
          if ((u32REG_SysRead(REG_SYS_FLASH_REMAP) & 0x0F) != JN516x::chipFlashSectors(CHIP_VERSION) - 1)
          {
            if (WPanOTAStartCallback)
              WPanOTAStartCallback();
            disabeApp();
            vAHI_SwReset();
          }
          break;
        case 0xFE:
          if (bAHI_FullFlashProgram(sNetwork.u32OtaStart, sizeof(sNetwork.u8OtaHeader), sNetwork.u8OtaHeader))
            vAHI_SwReset();
          break;
      }
    }
  }
}

void WPan::disabeApp(void)
{
  uint8 u8Buf[32];
  if (bAHI_FullFlashRead(0, sizeof(u8Buf), u8Buf))
  {
#if defined(JN5179)
    ((JN516x::tsFlashHeader_JN5179 *)u8Buf)->sBIR.status = 0;
#else
    ((JN516x::tsFlashHeader *)u8Buf)->sBIR.status = 0;
#endif
    bAHI_FullFlashProgram(0, sizeof(u8Buf), u8Buf);
  }
}

#pragma GCC diagnostic pop

bool WPan::FlashInit(void)
{
  return bAHI_FlashInit(E_FL_CHIP_AUTO, 0);
}
