/*
  wpan.h - Wireless Personal Area Network Library for NXP-JN516x (IEEE802.15.4)

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <MMAC.h>
#include <AppQueueApi.h>
#include "wpan_security.h"

#ifndef WPAN_DEFAULT_PANID
#define WPAN_DEFAULT_PANID 0x5160
#endif

#ifndef WPAN_DEFAULT_CHANNEL
#define WPAN_DEFAULT_CHANNEL CHANNEL_23
#endif

#ifndef WPAN_MAX_DEVICES
#define WPAN_MAX_DEVICES 128U
#endif

#ifndef WPAN_RSV_DEVICES
#define WPAN_RSV_ADDRESS 8U
#endif

#define WPAN_ADDR_BROADCAST   0xFFFFU

#if WPAN_SECURITY_ENABLE
#define WPAN_MAX_PAYLOAD_SIZE (114 - 1 - WPAN_SECURITY_AES_MODE)
#else
#define WPAN_MAX_PAYLOAD_SIZE (114 - 1)
#endif

extern void WPanRecieveCallback(uint16 u16Addr, const uint8 *pu8Data, uint8 u8Len, bool broadcast) __attribute__((weak));
extern void WPanTransmitCallback(uint8 u8Handle, uint8 u8Status) __attribute__((weak));
extern void WPanHWEventCallback(uint32 u32Device, uint32 u32ItemBitmap) __attribute__((weak));
extern void WPanOTAStartCallback(void) __attribute__((weak));

class WPan
{
  public:

    typedef enum
    {
      /* 868 MHz (Europe) */
      CHANNEL_0,  // 868.3
      /* 915 MHz (America, Australia) */
      CHANNEL_1,  // 906
      CHANNEL_2,  // 908
      CHANNEL_3,  // 910
      CHANNEL_4,  // 912
      CHANNEL_5,  // 914
      CHANNEL_6,  // 916
      CHANNEL_7,  // 918
      CHANNEL_8,  // 920
      CHANNEL_9,  // 922
      CHANNEL_10, // 924
      /* 2400 MHz (Worldwide) */
      CHANNEL_11, // 2405
      CHANNEL_12, // 2410
      CHANNEL_13, // 2415
      CHANNEL_14, // 2420
      CHANNEL_15, // 2425
      CHANNEL_16, // 2430
      CHANNEL_17, // 2435
      CHANNEL_18, // 2440
      CHANNEL_19, // 2445
      CHANNEL_20, // 2450
      CHANNEL_21, // 2455
      CHANNEL_22, // 2460
      CHANNEL_23, // 2465
      CHANNEL_24, // 2470
      CHANNEL_25, // 2475
      CHANNEL_26, // 2480
    } CHANNEL;

    static bool begin(uint16 u16Addr, uint16 u16PanId = WPAN_DEFAULT_PANID, CHANNEL u8Channel = WPAN_DEFAULT_CHANNEL);
    static void transmit(uint16 u16Addr, const uint8 *pu8Data, uint8 u8Len, uint8 u8Handle = 0);
    static void reTransmit(uint16 u16Addr);
    static void handle(void);
    static bool setShortAddress(uint16 u16Addr);
    static uint16 getShortAddress(void);
    static void getMacAddress(MAC_ExtAddr_s &addr);

    typedef enum
    {
      TXPOWER_0 = -32, // -32 dBm
      TXPOWER_1 = -20, // -20 dBm
      TXPOWER_2 =  -9, //  -9 dBm
      TXPOWER_3 =   0, //   0 dBm
#if E_AHI_ADC_SRC_ADC_5
      TXPOWER_4 =  10, //  10 dBm (JN5169)
#endif
    } TXPOWER;

    static void transmissionPower(TXPOWER txpower);

  protected:

    typedef struct __attribute__ ((__packed__))
    {
      uint8 u8Tx;
      uint8 u8Rx;
    } tsSeqNb;

    typedef struct transmit
    {
      struct transmit *next;
      MAC_McpsReqRsp_s data;
    } tsTransmit;

    /* Data related to associated end devices */
    typedef struct
    {
      uint16      u16Addr;
      uint16      u16PanId;
      tsSeqNb     sSeqNb[WPAN_MAX_DEVICES + WPAN_RSV_ADDRESS];
      uint32      u32OtaStart;
      uint32      u32OtaAddr;
      uint8       u8OtaErase;
      uint8       u8OtaHandle;
      uint8       u8OtaHeader[32];
      bool        bOtaValid;
      tsTransmit *psTransmit;
      tsTransmit  sTransmitData[4];
    } tsNetwork;

    static void vProcessIncomingMlme(MAC_MlmeDcfmInd_s *psMlmeInd);
    static void vProcessIncomingMcps(MAC_McpsDcfmInd_s *psMcpsInd);
    static void vHandleMcpsDataInd(MAC_RxFrameData_s& sFrame);
    static void vHandleMcpsDataDcfm(MAC_McpsCfmData_s& sDcfmData);
    static void vProcessIncomingHwEvent(AppQApiHwInd_s *psAHI_Ind);
    static tsTransmit *getTransmitData(uint16 u16Addr, bool alloc = true);
    static void removeTransmitData(tsTransmit *psTransmit);
    static tsSeqNb *getSequenceNumber(uint16 u16Addr);
    static void vSetMacAddr(MAC_Addr_s& sAddr, uint16 u16Addr);
    static void write32(uint8 *buf, uint32 data);
    static uint32 read32(const uint8 *buf);
    static void OTARecieveCallback(uint16 u16Addr, const uint8 *pu8Data, uint8 u8Len);
    static void OTATransmitCallback(uint8 u8Handle, uint8 u8Status);
    static void disabeApp(void);
    static bool FlashInit(void);

    /* Handles from the MAC */
    static tsNetwork sNetwork;
};
