/*
  main.cpp - Garage Shutter Control Library for JN516X

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "device.h"
#include "shutter.h"

#define EXTERNAL_ALARM_ID    0x01
#define CONNECT_ACTIVE_TIME  4
#define PAIRING_MODE_TIME    5000
#define MANTENANCE_MODE_TIME 10000

enum
{
  PAIRING_DEVICE,
  PAIRING_REGISTER,
  PAIRING_IGNORE,
};

static uint16_t  peer;
static uint16_t  self;
static command_t packet;
static uint8_t   active;
static uint8_t   alarm;
static uint8_t   pairing_mode;
static uint8_t   maintenance;
static bool      select_mode;
static uint32_t  select_start;
static uint32_t  clock;
static Shutter   shutter;
static macaddr_t macaddr;

static const macaddr_t MACNONE = {};

static void event(uint8_t event, uint8_t id)
{
  switch (event)
  {
    case Shutter::EVENT_PAUSE:
      if (id != EXTERNAL_ALARM_ID)
        alarm |= 0x81;
      break;
    case Shutter::EVENT_RESUME:
      if (id != EXTERNAL_ALARM_ID)
        alarm &= ~0x01;
      break;
  }
}

static void pairing(uint16_t u16Addr)
{
  bool ack = false;
  switch (pairing_mode)
  {
    case PAIRING_DEVICE:
      if (memcmp(&packet.args.pairing.mac, &macaddr, sizeof(macaddr)) == 0)
      {
        switch(packet.args.pairing.ctrl)
        {
          case PAIRING_DOWN  : shutter.down();    break;
          case PAIRING_UP    : shutter.up();      break;
          case PAIRING_AUTOUP: shutter.auto_up(); break;
          case PAIRING_UPDOWN:
            if (shutter.status() >= 50)
              shutter.up();
            else
              shutter.down();
            break;
        }
        ack = true;
      }
      break;
    case PAIRING_REGISTER:
      if (memcmp(&packet.args.pairing.mac, &MACNONE, sizeof(MACNONE)) == 0)
      {
        // register
        memcpy(&packet.args.pairing.mac, &macaddr, sizeof(macaddr));
        pairing_mode = PAIRING_IGNORE;
        ack = true;
      }
      else if (memcmp(&packet.args.pairing.mac, &macaddr, sizeof(macaddr)) == 0)
      {
        // unregister
        memcpy(&packet.args.pairing.mac, &MACNONE, sizeof(macaddr));
        pairing_mode = PAIRING_IGNORE;
        ack = true;
      }
      break;
  }
  if (ack)
  {
    WPan::transmit(u16Addr, (uint8_t *)&packet,
      offsetof(__typeof__(packet), args) + sizeof(packet.args.pairing));
  }
}

static void acknowledge(uint16_t u16Addr, uint8_t err)
{
  packet.args.data[0] = err;
  WPan::transmit(u16Addr, (uint8_t *)&packet,
    offsetof(__typeof__(packet), args) + sizeof(packet.args.data[0]));
}

void WPanRecieveCallback(uint16 u16Addr, const uint8 *pu8Data, uint8 u8Len, bool broadcast)
{
  if (u8Len)
  {
    memcpy(&packet, pu8Data, u8Len);
    if (broadcast)
    {
      switch (packet.command)
      {
        case BRIDGE_CONNECT:
          if (memcmp(&packet.args.address.mac, &macaddr, sizeof(macaddr)) == 0)
          {
            uint16_t addr = (((uint16_t)packet.args.address.addrH) << 8) | packet.args.address.addrL;
            if ((WPan::setShortAddress(addr)))
            {
              peer = u16Addr;
              if (self != addr)
              {
                self = addr;
                eeprom_erase(EEPROM_CONFIG_OFFSET, sizeof(self));
                eeprom_write(EEPROM_CONFIG_OFFSET, (uint8_t *)&self, sizeof(self));
              }
              acknowledge(u16Addr, 0);
              active = CONNECT_ACTIVE_TIME;
            }
          }
          break;
        case DEVICE_PAIRING:
          pairing(u16Addr);
          break;
      }
    }
    else if (u16Addr == peer)
    {
      switch (packet.command)
      {
        case SHUTTER_SETTING:
          shutter.setting(&packet.args.setting);
          acknowledge(u16Addr, 0);
          break;
        case SHUTTER_STOP:
          shutter.stop();
          acknowledge(u16Addr, 0);
          break;
        case SHUTTER_DOWN:
          shutter.down();
          acknowledge(u16Addr, 0);
          break;
        case SHUTTER_UP:
          shutter.up();
          acknowledge(u16Addr, 0);
          break;
        case SHUTTER_AUTO_UP:
          shutter.auto_up();
          acknowledge(u16Addr, 0);
          break;
        case SHUTTER_PAUSE:
          shutter.pause(EXTERNAL_ALARM_ID);
          acknowledge(u16Addr, 0);
          break;
        case SHUTTER_RESUME:
          shutter.resume(EXTERNAL_ALARM_ID);
          acknowledge(u16Addr, 0);
          break;
        case SHUTTER_ALARM:
          shutter.alarm_control(packet.args.data[0]);
          acknowledge(u16Addr, 0);
          break;
        case BRIDGE_POLLING:
          acknowledge(u16Addr, shutter.status() | (alarm & 0x80));
          if ((alarm & 0x01) == 0)
            alarm = 0;
          break;
        case DEVICE_RESET:
          acknowledge(u16Addr, 0);
          led_toggle(0, 100);
          delay(1000);
          reset();
          break;
        default:
          return;
      }
      active = CONNECT_ACTIVE_TIME;
    }
  }
}

void setup()
{
#if 0
  eeprom_erase();
#endif
  //
  // Init Shutter Controller
  //
  shutter.begin();
  shutter.callback(event);
  //
  // Read Device Address
  //
  eeprom_read(EEPROM_CONFIG_OFFSET, (uint8_t *)&self, sizeof(self));
  //
  // Init WPan
  //
  WPan::begin(self);
  //
  // Read MAC Address
  //
  MAC_ExtAddr_s mac;
  WPan::getMacAddress(mac);
  macaddr.data[7] = mac.u32L >> 0;
  macaddr.data[6] = mac.u32L >> 8;
  macaddr.data[5] = mac.u32L >> 16;
  macaddr.data[4] = mac.u32L >> 24;
  macaddr.data[3] = mac.u32H >> 0;
  macaddr.data[2] = mac.u32H >> 8;
  macaddr.data[1] = mac.u32H >> 16;
  macaddr.data[0] = mac.u32H >> 24;
  //
  // Start Watch Dog
  //
#ifdef WATCHDOG_ENABLED
  wdt_start();
#endif
  //
  // Save System Clock
  //
  clock = millis();
}

void loop()
{
  uint32_t now = millis();
  //
  // control led
  //
  uint32_t elaps  = now - clock;
  uint32_t elaps2 = select_mode ? now - select_start : 0;
  if (maintenance ? elaps2 < MANTENANCE_MODE_TIME : elaps2 >= MANTENANCE_MODE_TIME)
  {
    if (elaps >= 100)
    {
      clock += 100;
      led_on();
      led_off(0, 10);
    }
  }
  else if (!maintenance
    && (pairing_mode != PAIRING_DEVICE ? elaps2 < PAIRING_MODE_TIME : elaps2 >= PAIRING_MODE_TIME))
  {
    if (elaps >= 500)
    {
      clock += 500;
      led_on();
      led_off(0, 10);
    }
  }
  else if (elaps >= 1000)
  {
    clock += 1000;
    if (active)
      active--;
    if (alarm & 0x01)
    {
      led_off();
      led_on(0, 10);
    }
    else if (!active)
    {
      led_on();
      led_off(0, 10);
    }
    else
    {
      led_toggle();
    }
  }
  //
  // select mode (pairing and maintenance)
  //
  if (shutter.stop_button())
  {
    pairing_mode = PAIRING_DEVICE;
    if (!shutter.down_button() || !shutter.up_button())
      select_mode = false;
    else if (!select_mode)
    {
      select_mode = true;
      select_start = now;
    }
  }
  else if (select_mode)
  {
    select_mode = false;
    if (elaps2 >= MANTENANCE_MODE_TIME)
      maintenance = !maintenance;
    else if (elaps2 >= PAIRING_MODE_TIME)
      pairing_mode = PAIRING_REGISTER;
  }
  //
  // handle shutter controller
  //
  shutter.handle(maintenance);
}
