/*
  shutter.cpp - Garage Shutter Control Library for JN516X

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <string.h>
#include "device.h"
#include "shutter.h"

#define GPIO_ALARM_EN Gpio::DIO5
#define GPIO_UP_OUT   Gpio::DIO13
#define GPIO_STOP_OUT Gpio::DIO14
#define GPIO_DOWN_OUT Gpio::DIO11
#define GPIO_UP_BTN   Gpio::DIO18
#define GPIO_STOP_BTN Gpio::DIO19
#define GPIO_DOWN_BTN Gpio::DIO4

#define GPIO_ALARM_ID   0x80
#define CHATTERING_TIME 10 // ms
#define _BV(b)          (1 << (b))

enum
{
  STOP_BTN = 0,
  DOWN_BTN = 1,
  UP_BTN   = 2,
  ALARM    = 7,
};

enum
{
  CLOSE    = 0,
  OPEN     = 1,
  PAUSE    = 7,
};

enum
{
  STOP,
  DOWN,
  UP,
};

void Shutter::event(EVENT event, uint8_t id)
{
  if (_callback)
    _callback(event, id);
}

void Shutter::callback(callback_t cb)
{
  _callback = cb;
}

void Shutter::opend(void)
{
  _t_elaps   = 0;
  _t_counter = _t_open = _setting.open_time;
  _status   |= _BV(OPEN);
  _operation = STOP;
  event(EVENT_OPENED, 0);
}

void Shutter::closed(void)
{
  _t_elaps   = 0;
  _t_counter = 0;
  _status   |= _BV(CLOSE);
  _operation = _shutter = STOP; // follow shutter control
  event(EVENT_CLOSED, 0);
}

void Shutter::output(uint8_t pin, uint8_t val)
{
  // Since GPIO_STOP_OUT is for both input and output,
  // it is handled by changing the pin mode.
  digitalWrite(pin, LOW);
  pinMode(pin, val ? INPUT_PULLUP : OUTPUT);
}

void Shutter::control(void)
{
  //
  // calculate time span
  //
  uint32_t now = millis();
  uint32_t elaps = now - _t_start;
  _t_start = now;
  //
  // Prevention of forgetting to close (3 minutes or 12 hours)
  //
  if (!_maintenance)
  {
    if ((_t_elaps += elaps) >= ((_closing && (_status & _BV(CLOSE)))
      ?       3 * 60 * (uint32_t)1000
      : 12 * 60 * 60 * (uint32_t)1000))
    {
      _t_elaps = 0;
      _operation = DOWN;
    }
  }
  //
  // update open/close timer
  //
  if (!(_status & _BV(OPEN)) && (_shutter == UP))
  {
    if ((_t_counter += elaps) >= _t_open)
      opend();
  }
  else if (!(_status & _BV(CLOSE)) && (_shutter == DOWN))
  {
    if ((int32_t)(_t_counter -= elaps) <= 0)
      closed();
  }
  //
  // auto close control
  //
  if (_t_aclose)
  {
    if ((int32_t)(_t_aclose -= elaps) <= 0)
    {
      _t_aclose  = 0;
      _t_elaps   = 0;
      _operation = DOWN;
      event(EVENT_AUTO_TIMEUP, 0);
    }
  }
  //
  // resume control
  //
  if ((_status & _BV(PAUSE)) && _t_resume)
  {
    if ((int32_t)(_t_resume -= elaps) <= 0)
    {
      _t_resume = 0;
      _status &= ~_BV(PAUSE);
      event(EVENT_RESUME_TIMEUP, 0);
    }
  }
  //
  // pulse output control
  //
  if (_pulse_on[0] != 0xFF)
  {
    switch (_io_step)
    {
      case 0:
        if (now - _t_pulse >= _setting.pulse_time)
        {
          output(_pulse_on[0], HIGH);
          _pulse_on[0] = _pulse_on[1];
          _pulse_on[1] = 0xFF;
          if (_pulse_on[0] != 0xFF)
          {
            _io_step = 1;
            _t_pulse = now;
          }
        }
        break;
      case 1:
        if (now - _t_pulse >= _setting.pulse_interval)
        {
          output(_pulse_on[0], LOW);
          _io_step = 0;
          _t_pulse = now;
          switch (_pulse_on[0])
          {
            case GPIO_DOWN_OUT:
              _shutter = DOWN;
              _status &= ~(_BV(CLOSE) | _BV(OPEN));
              break;
            case GPIO_UP_OUT:
              _shutter = UP;
              _status &= ~(_BV(CLOSE) | _BV(OPEN));
              break;
          }
        }
        break;
    }
  }
}

void Shutter::stop0(bool alarm)
{
  alarm_control(alarm);
  if (_pulse_on[0] != 0xFF)
    output(_pulse_on[0], HIGH);
  _pulse_on[0] = GPIO_STOP_OUT;
  _pulse_on[1] = 0xFF;
  output(_pulse_on[0], LOW);
  _closing = false;
  _shutter = STOP;
  _io_step = 0;
  _t_pulse = millis();
}

void Shutter::down0(void)
{
  stop0(true);
  _pulse_on[1] = GPIO_DOWN_OUT;
  _closing = true;
}

void Shutter::up0(void)
{
  stop0(false);
  _pulse_on[1] = GPIO_UP_OUT;
}

void Shutter::stop(void)
{
  if (_maintenance)
    return;
  event(EVENT_STOP, 0);
  stop0(true);
  _operation = STOP;
  _t_elaps   = 0;
  _t_aclose  = 0;
}

void Shutter::down(void)
{
  if (_maintenance)
    return;
  event(EVENT_DOWN, 0);
  if ((_operation != DOWN))// && (_sensors & (_BV(CLOSED) | _BV(OPEN))) != _BV(CLOSE)))
  {
    if (!(_status & _BV(PAUSE)))
      down0();
    _operation = DOWN;
  }
  _t_elaps  = 0;
  _t_aclose = 0;
}

void Shutter::up(uint32_t autodw)
{
  if (_maintenance)
    return;
  event(autodw ? EVENT_AUTO_UP : EVENT_UP, 0);
  if ((_operation != UP) && ((_status & (_BV(CLOSE) | _BV(OPEN))) != _BV(OPEN)))
  {
    up0();
    _operation = UP;
  }
  _t_elaps  = 0;
  _t_aclose = autodw;
}

void Shutter::auto_up(void)
{
  up(_setting.autodw_time);
}

void Shutter::pause(uint8_t id)
{
  if (_maintenance)
    return;
  event(EVENT_PAUSE, id);
  if (id & ~_pauses)
  {
    stop0(true);
    _pauses  |= id;
    _status  |= _BV(PAUSE);
    _t_resume = 0;
  }
  _t_elaps = 0;
}

void Shutter::resume(uint8_t id)
{
  if (_maintenance)
    return;
  event(EVENT_RESUME, id);
  if (id & _pauses)
  {
    if ((_pauses &= ~id) == 0)
      _t_resume = _setting.resume_time;
  }
  _t_elaps = 0;
}

uint8_t Shutter::status(void)
{
  return 100 - (_t_counter * 100) / _t_open;
}

uint8_t Shutter::edge_state(uint8_t id)
{
  return _pin_state & _BV(id) ? HIGH : LOW;
}

void Shutter::edge_init(uint8_t id)
{
  id = _BV(id);
  _pin_flags |= id;
  _pin_state |= id;
}

uint8_t Shutter::edge_detect(uint8_t id, uint8_t pin)
{
  if (_pulse_on[0] != pin)
  {
    uint32_t now = millis();
    uint8_t  val;
    id  = _BV(id);
    val = digitalRead(pin) ? id : 0;
    if ((_pin_flags ^ val) & id)
    {
      _pin_flags ^= id;
      _pin_timer  = now;
    }
    else if (((_pin_state ^ _pin_flags) & id) && ((now - _pin_timer) >= CHATTERING_TIME))
    {
      return (_pin_state ^= id) & id ? HIGH : LOW;
    }
  }
  return 0xFF;
}

bool Shutter::stop_button(void)
{
  return edge_state(STOP_BTN) == LOW;
}

bool Shutter::down_button(void)
{
  return edge_state(DOWN_BTN) == LOW;
}

bool Shutter::up_button(void)
{
  return edge_state(UP_BTN) == LOW;
}

void Shutter::setting(setting_t *opt)
{
  if (memcmp(&_setting, opt, sizeof(_setting)) != 0)
  {
    _t_open = _setting.open_init ? _setting.open_init : _setting.open_time;
    memcpy(&_setting, opt, sizeof(_setting));
    alarm_control(true);
    eeprom_erase(EEPROM_SETTING_OFFSET, sizeof(_setting));
    eeprom_write(EEPROM_SETTING_OFFSET, (uint8_t *)&_setting, sizeof(_setting));
  }
}

void Shutter::alarm_control(bool enable)
{
  if (_maintenance)
    return;
  //
  // alarm control
  //
  digitalWrite(GPIO_ALARM_EN, enable && !_setting.alarm_disable ? LOW : HIGH);
}

void Shutter::begin(void)
{
  //
  // init variables
  //
  _maintenance = false;
  _operation   = STOP;
  _shutter     = STOP;
  _status      = 0;
  _pauses      = 0;
  _closing     = false;
  _io_step     = 0;
  _t_open      = 0;
  _t_start     = 0;
  _t_elaps     = 0;
  _t_counter   = 0;
  _t_aclose    = 0;
  _t_resume    = 0;
  _t_pulse     = 0;
  _pulse_on[0] = 0xFF;
  _pulse_on[1] = 0xFF;
  _pin_timer   = 0;
  _pin_flags   = 0;
  _pin_state   = 0;
  //
  // init gpio
  //
  pinMode(GPIO_STOP_BTN, INPUT_PULLUP);
  pinMode(GPIO_DOWN_BTN, INPUT_PULLUP);
  pinMode(GPIO_UP_BTN  , INPUT_PULLUP);
  pinMode(GPIO_STOP_OUT, INPUT_PULLUP);
  pinMode(GPIO_DOWN_OUT, INPUT_PULLUP);
  pinMode(GPIO_UP_OUT  , INPUT_PULLUP);
  pinMode(GPIO_ALARM_EN, OUTPUT);
  digitalWrite(GPIO_ALARM_EN, HIGH);
  //
  // init edge sense
  //
  edge_init(STOP_BTN);
  edge_init(UP_BTN  );
  edge_init(DOWN_BTN);
  edge_init(ALARM   );
  //
  // read eeprom data
  //
  eeprom_read(EEPROM_SETTING_OFFSET, (uint8_t *)&_setting, sizeof(_setting));
  if (_setting.pulse_time == 0)
  {
    _setting.pulse_time     = 1000;
    _setting.pulse_interval = 100;
    _setting.open_init      = 0;
    _setting.open_time      = 28000;
    _setting.resume_time    = 15000;
    _setting.autodw_time    = 30000;
    _setting.alarm_disable  = 0;
  }
  _t_open = _setting.open_init ? _setting.open_init : _setting.open_time;
  //
  // set start time
  //
  _t_start = millis();
}

void Shutter::handle(uint8_t maintenance)
{
  if (_maintenance != maintenance)
  {
    output(GPIO_STOP_OUT, HIGH);
    output(GPIO_DOWN_OUT, HIGH);
    output(GPIO_UP_OUT  , HIGH);
    delay(_setting.pulse_interval);
    stop();
  }
  if ((_maintenance = maintenance))
  {
    //
    // maintenance mode
    //
    switch (edge_detect(STOP_BTN, GPIO_STOP_BTN))
    {
      case LOW : output(GPIO_STOP_OUT, LOW ); break;
      case HIGH: output(GPIO_STOP_OUT, HIGH); break;
    }
    switch (edge_detect(DOWN_BTN, GPIO_DOWN_BTN))
    {
      case LOW : output(GPIO_DOWN_OUT, LOW ); break;
      case HIGH: output(GPIO_DOWN_OUT, HIGH); break;
    }
    switch (edge_detect(UP_BTN, GPIO_UP_BTN))
    {
      case LOW : output(GPIO_UP_OUT, LOW ); break;
      case HIGH: output(GPIO_UP_OUT, HIGH); break;
    }
  }
  else
  {
    //
    // check manual control button
    //
    if (edge_detect(STOP_BTN, GPIO_STOP_BTN) == LOW)
    {
	    stop();
    }
    else if (edge_detect(DOWN_BTN, GPIO_DOWN_BTN) == LOW)
    {
      if (!stop_button()) down();
    }
    else if (edge_detect(UP_BTN, GPIO_UP_BTN) == LOW)
    {
      if (!stop_button()) up();
    }
    //
    // check alarm sensor
    //
    switch (edge_detect(ALARM, GPIO_STOP_OUT))
    {
      case LOW : pause (GPIO_ALARM_ID); break;
      case HIGH: resume(GPIO_ALARM_ID); break;
    }
    //
    // resume control
    //
    if ((_pulse_on[0] == 0xFF) && (_shutter != _operation))
    {
      switch (_operation)
      {
        case STOP:
          stop0(true);
          break;
        case DOWN:
      	  if (_status & _BV(PAUSE))
      	    break;
         	if (_pulse_on[1] != GPIO_DOWN_OUT)
            down0();
          break;
        case UP:
          if (_pulse_on[1] != GPIO_UP_OUT)
            up0();
          break;
      }
    }
  }
  //
  // timer control
  //
  control();
}
