/*
  shutter.h - Garage Shutter Control Library for JN516X

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <stdint.h>
#include <stdbool.h>
#include "hostdef.h"

#define INPUT                  Gpio::INPUT
#define INPUT_PULLUP           Gpio::INPUT_PULLUP
#define OUTPUT                 Gpio::OUTPUT
#define pinMode(pin, mode)     Gpio::pinMode((Gpio::PIN)(pin), mode)
#define digitalRead(pin)       Gpio::digitalRead((Gpio::PIN)(pin))
#define digitalWrite(pin, val) Gpio::digitalWrite((Gpio::PIN)(pin), val)
#define millis()               System::millis()
#define delay(ms)              System::delay(ms)
#define reset()                System::reset()
#define eeprom_read            EEPRom::read
#define eeprom_write           EEPRom::write
#define eeprom_erase           EEPRom::erase
#define led_on                 Led::on
#define led_off                Led::off
#define led_toggle             Led::toggle
#define wpan_begin             WPan::begin
#define wpan_getMacAddress     WPan::getMacAddress
#define wpan_setShortAddress   WPan::setShortAddress
#define wpan_transmit          WPan::transmit
#define wdt_start()            Watchdog::start(Watchdog::WDTO_1032MS)

#define EEPROM_CONFIG_OFFSET   0
#define EEPROM_SETTING_OFFSET  64

class Shutter
{
  public:

    typedef enum
    {
      EVENT_STOP,
      EVENT_DOWN,
      EVENT_UP,
      EVENT_AUTO_UP,
      EVENT_AUTO_TIMEUP,
      EVENT_PAUSE,
      EVENT_RESUME,
      EVENT_RESUME_TIMEUP,
      EVENT_OPENED,
      EVENT_CLOSED,
    } EVENT;;

    typedef void (*callback_t)(uint8_t event, uint8_t id);

    void opend(void);
    void closed(void);
    void stop(void);
    void down(void);
    void up(uint32_t autodw = 0);
    void auto_up(void);
    void pause(uint8_t id);
    void resume(uint8_t id);
    void setting(setting_t *opt);
    void alarm_control(bool enable);
    void begin(void);
    void handle(uint8_t maintenance);
    uint8_t status(void);
    void callback(callback_t cb);
    bool stop_button(void);
    bool down_button(void);
    bool up_button(void);

  private:

    void event(EVENT event, uint8_t id);
    void output(uint8_t pin, uint8_t val);
    void control(void);
    void stop0(bool alarm);
    void down0(void);
    void up0(void);
    uint8_t edge_state(uint8_t id);
    void edge_init(uint8_t id);
    uint8_t edge_detect(uint8_t id, uint8_t pin);

    setting_t  _setting;
    bool       _maintenance;
    uint8_t    _operation;
    uint8_t    _shutter;
    uint8_t    _status;
    uint8_t    _pauses;
    uint8_t    _percent;
    bool       _closing;
    uint8_t    _io_step;
    uint32_t   _t_open;
    uint32_t   _t_start;
    uint32_t   _t_elaps;
    uint32_t   _t_counter;
    uint32_t   _t_aclose;
    uint32_t   _t_resume;
    uint32_t   _t_pulse;
    uint8_t    _pulse_on[2];
    uint32_t   _pin_timer;
    uint8_t    _pin_flags;
    uint8_t    _pin_state;
    callback_t _callback;
};
